#!/usr/bin/env python
# -*- coding: utf-8 -*-

import inspect
from typing import Any, Callable, Generic, List, TypeVar, overload

T = TypeVar("T")
U = TypeVar("U")


def identity(x: T) -> T:
    """Identity function placeholder."""
    return x


class Compose(Generic[T, U]):
    @overload
    def __init__(self) -> None:
        ...

    @overload
    def __init__(
        self,
        fn0: Callable[[T], U],
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        fn0: Callable[[T], Any],
        fn1: Callable[[Any], U],
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        fn0: Callable[[T], Any],
        fn1: Callable[[Any], Any],
        fn2: Callable[[Any], U],
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        fn0: Callable[[T], Any],
        fn1: Callable[[Any], Any],
        fn2: Callable[[Any], Any],
        fn3: Callable[[Any], U],
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        fn0: Callable[[T], Any],
        fn1: Callable[[Any], Any],
        fn2: Callable[[Any], Any],
        fn3: Callable[[Any], Any],
        fn4: Callable[[Any], U],
        /,
    ) -> None:
        ...

    @overload
    def __init__(self, *fns: Callable) -> None:
        ...

    def __init__(self, *fns: Callable) -> None:
        super().__init__()
        self.fns = fns

    def __call__(self, x: T) -> U:
        for fn in self.fns:
            x = fn(x)
        return x  # type: ignore

    def __getitem__(self, idx: int, /) -> Callable[[Any], Any]:
        return self.fns[idx]

    def __len__(self) -> int:
        return len(self.fns)


compose = Compose  # type: ignore


def get_argnames(fn: Callable) -> List[str]:
    """Get arguments names of a method, function or callable object."""
    if inspect.ismethod(fn):
        # If method, remove 'self' arg
        argnames = fn.__code__.co_varnames[1:]  # type: ignore
    elif inspect.isfunction(fn):
        argnames = fn.__code__.co_varnames
    else:
        argnames = fn.__call__.__code__.co_varnames  # type: ignore

    argnames = list(argnames)
    return argnames


def filter_and_call(fn: Callable[..., T], **kwargs: Any) -> T:
    """Filter kwargs with function arg names and call function."""
    argnames = get_argnames(fn)
    kwargs_filtered = {
        name: value for name, value in kwargs.items() if name in argnames
    }
    return fn(**kwargs_filtered)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch
from torch import Tensor

from torchoutil.nn.functional.indices import (
    get_inverse_perm,
    get_perm_indices,
    randperm_diff,
)


class TestRandpermDiff(TestCase):
    def test_example_1(self) -> None:
        for i in range(2, 10):
            perm = randperm_diff(i)
            arange = torch.arange(0, i, device=perm.device)

            assert isinstance(perm, Tensor)
            assert perm.shape == arange.shape
            assert not perm.eq(arange).any(), f"{perm}; {arange}"


class TestGetInversePerm(TestCase):
    def test_example_1(self) -> None:
        x = torch.as_tensor([2, 4, 8, 10])
        indices = torch.randperm(len(x))
        x = x[indices]
        # x is now shuffled, to get back the original order we need the indices
        inv_indices = get_inverse_perm(indices)
        x_reordered = x[inv_indices]

        assert torch.equal(x_reordered, torch.as_tensor([2, 4, 8, 10]))

    def test_examples_random(self) -> None:
        n_steps = 5
        for _ in range(n_steps):
            size = int(torch.randint(0, 100, ()).item())
            perm = torch.randperm(size)
            inv_perm = get_inverse_perm(perm)

            x1 = torch.rand(size, 10)
            x2 = x1[perm]
            x3 = x2[inv_perm]

            assert torch.equal(x1, x3)


class TestGetPermIndices(TestCase):
    def test_example_1(self) -> None:
        n_steps = 10
        for _ in range(n_steps):
            size = int(torch.randint(1, 100, ()).item())

            x1 = torch.randperm(size)
            perm = torch.randperm(size)
            x2 = x1[perm]
            perm_result = get_perm_indices(x1, x2)

            assert torch.equal(x1, x2[perm_result]), f"{x1=}, {x2=}"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from typing import Dict, Iterable, List, Optional, Tuple, Union

from torch import nn
from torch.nn.parameter import Parameter
from torch.optim.optimizer import Optimizer

pylog = logging.getLogger(__name__)


def get_lr(optim: Optimizer, idx: int = 0, key: str = "lr") -> float:
    """
    Get the learning rate of the first group of an optimizer.

    Args:
        optim: The optimizer to get.
        idx: The group index of the learning rate in the optimizer. defaults to 0.
    """
    return get_lrs(optim, key)[idx]


def get_lrs(optim: Optimizer, key: str = "lr") -> List[float]:
    """
    Get the learning rates in all groups of an optimizer.

    Args:
        optim: The optimizer to get.
    """
    return [group[key] for group in optim.param_groups]


def create_params_groups_bias(
    model: Union[nn.Module, Iterable[Tuple[str, Parameter]]],
    weight_decay: float,
    skip_list: Optional[Iterable[str]] = (),
    verbose: int = 2,
) -> List[Dict[str, Union[List[Parameter], float]]]:
    if isinstance(model, nn.Module):
        params = model.named_parameters()
    else:
        params = model
    del model

    decay: List[Parameter] = []
    no_decay: List[Parameter] = []

    if skip_list is None:
        skip_list = {}
    else:
        skip_list = dict.fromkeys(skip_list)

    for name, param in params:
        if not param.requires_grad:
            continue
        if len(param.shape) == 1 or name.endswith(".bias") or name in skip_list:
            no_decay.append(param)
            if verbose >= 2:
                pylog.debug(f"No wd for {name}")
        else:
            decay.append(param)

    return [
        {"params": decay, "weight_decay": weight_decay},
        {"params": no_decay, "weight_decay": 0.0},
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
from difflib import SequenceMatcher
from typing import Callable, Iterable


def sequence_matcher_ratio(a: str, b: str) -> float:
    return SequenceMatcher(None, a, b).ratio()


def find_closest_in_list(
    x: str,
    lst: Iterable[str],
    sim_fn: Callable[[str, str], float] = sequence_matcher_ratio,
    higher_is_closer: bool = True,
) -> str:
    best_sim = -int(higher_is_closer) * math.inf

    for elt in lst:
        sim = sim_fn(x, elt)
        if (higher_is_closer and best_sim < sim) or (
            not higher_is_closer and best_sim > sim
        ):
            best_sim = sim
            closest = elt

    if math.isinf(best_sim):
        msg = f"Invalid argument {lst=}. (expected non-empty iterable of strings)"
        raise ValueError(msg)

    return closest



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv
import io
from csv import DictReader, DictWriter
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    TypeVar,
    Union,
    overload,
)

from torchoutil.pyoutil.collections import (
    dict_list_to_list_dict,
    list_dict_to_dict_list,
)

T = TypeVar("T")

ORIENT_VALUES = ("list", "dict")


def to_csv(
    data: Union[Iterable[Mapping[str, Any]], Mapping[str, Iterable[Any]]],
    fpath: Union[str, Path, None] = None,
    *,
    overwrite: bool = True,
    make_parents: bool = True,
    header: bool = True,
    align_content: bool = False,
    **csv_writer_kwds,
) -> str:
    """Dump content to CSV format."""
    if fpath is not None:
        fpath = Path(fpath).resolve().expanduser()
        if not overwrite and fpath.exists():
            raise FileExistsError(f"File {fpath} already exists.")
        elif make_parents:
            fpath.parent.mkdir(parents=True, exist_ok=True)

    if isinstance(data, Mapping):
        data_lst = dict_list_to_list_dict(data)  # type: ignore
    else:
        data_lst = list(data)
    del data

    if header:
        writer_cls = DictWriter
        if len(data_lst) == 0:
            fieldnames = []
        else:
            fieldnames = [str(k) for k in data_lst[0].keys()]
    else:
        writer_cls = csv.writer
        fieldnames = list(range(len(next(iter(data_lst)))))

    if align_content:
        old_fieldnames = fieldnames
        data_lst = _stringify(data_lst)
        fieldnames = _stringify(fieldnames)
        max_num_chars = {
            k: max(max(len(data_i[k]) for data_i in data_lst), len(k)) + 1
            for k in fieldnames
        }

        fieldnames = [f"{{:^{max_num_chars[k]}s}}".format(k) for k in fieldnames]
        old_to_new_fieldnames = dict(zip(old_fieldnames, fieldnames))

        data_lst = [
            {
                old_to_new_fieldnames[k]: f"{{:^{max_num_chars[k]}s}}".format(v)
                for k, v in data_i.items()
            }
            for data_i in data_lst
        ]

    if header:
        csv_writer_kwds["fieldnames"] = fieldnames

    file = io.StringIO()
    writer = writer_cls(file, **csv_writer_kwds)
    if isinstance(writer, DictWriter):
        writer.writeheader()
    writer.writerows(data_lst)  # type: ignore
    content = file.getvalue()
    file.close()

    if fpath is not None:
        fpath.write_text(content)

    return content


@overload
def load_csv(
    fpath: Union[str, Path],
    /,
    *,
    orient: Literal["dict"],
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    **csv_reader_kwds,
) -> Dict[str, List[Any]]:
    ...


@overload
def load_csv(
    fpath: Union[str, Path],
    /,
    *,
    orient: Literal["list"] = "list",
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    **csv_reader_kwds,
) -> List[Dict[str, Any]]:
    ...


def load_csv(
    fpath: Union[str, Path],
    /,
    *,
    orient: Literal["list", "dict"] = "list",
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    **csv_reader_kwds,
) -> Union[List[Dict[str, Any]], Dict[str, List[Any]]]:
    """Load content from csv filepath."""
    if header:
        reader_cls = DictReader
    else:
        reader_cls = csv.reader

    with open(fpath, "r") as file:
        reader = reader_cls(file, **csv_reader_kwds)
        raw_data_lst = list(reader)

    data_lst: List[Dict[str, Any]]
    if header:
        data_lst = raw_data_lst  # type: ignore
    else:
        data_lst = [
            {str(j): data_ij for j, data_ij in enumerate(data_i)}
            for data_i in raw_data_lst
        ]
    del raw_data_lst

    if comment_start is not None:
        data_lst = [
            line
            for line in data_lst
            if not next(iter(line.values())).startswith(comment_start)
        ]

    if strip_content:
        data_lst = [
            {k.strip(): v.strip() for k, v in data_i.items()} for data_i in data_lst
        ]

    if orient == "dict":
        result = list_dict_to_dict_list(data_lst, key_mode="same")  # type: ignore
    elif orient == "list":
        result = data_lst
    else:
        msg = f"Invalid argument {orient=}. (expected one of {ORIENT_VALUES})"
        raise ValueError(msg)

    return result  # type: ignore


def _stringify(x: Any) -> Any:
    if isinstance(x, str):
        return x
    elif isinstance(x, dict):
        return {_stringify(k): _stringify(v) for k, v in x.items()}  # type: ignore
    elif isinstance(x, (list, tuple, set, frozenset)):
        return type(x)(_stringify(xi) for xi in x)
    else:
        return str(x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Helper functions for conversion between classes indices, onehot, names and probabilities for multiclass classification.
"""

from typing import (
    Any,
    Callable,
    Hashable,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

import torch
from torch import Tensor
from torch.nn import functional as F

from torchoutil.core.get import DeviceLike, DTypeLike, get_device, get_dtype
from torchoutil.extras.numpy import np
from torchoutil.nn.functional.others import nelement, to_item
from torchoutil.pyoutil.logging import warn_once
from torchoutil.types import LongTensor, is_number_like
from torchoutil.types._typing import TensorLike

T_Name = TypeVar("T_Name", bound=Hashable)


def index_to_onehot(
    index: Union[Sequence[int], TensorLike, Sequence],
    num_classes: int,
    *,
    padding_idx: Optional[int] = None,
    device: DeviceLike = None,
    dtype: DTypeLike = torch.bool,
) -> Tensor:
    """Convert indices of labels to onehot boolean encoding for **multiclass** classification.

    Args:
        indices: List label indices.
            Can be a nested list of indices, but it should be convertible to Tensor.
        num_classes: Number maximal of unique classes.
        padding_idx: Class index to ignore. Output will contains only zeroes for this value. defaults to None.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    device = get_device(device)
    dtype = get_dtype(dtype)
    index = torch.as_tensor(index, device=device, dtype=torch.long)

    if padding_idx is not None:
        mask = index == padding_idx
        index = torch.where(mask, num_classes, index)
        num_classes += 1

    if index.nelement() > 0 and not (0 <= index.min() <= index.max() < num_classes):
        msg = f"Invalid argument {index=}. (expected 0 <= min={index.min()} <= max={index.max()} < {num_classes=})"
        raise ValueError(msg)

    onehot: Tensor = F.one_hot(index, num_classes)
    onehot = onehot.to(dtype=dtype)

    if padding_idx is not None:
        onehot = onehot[..., :-1]

    return onehot


def index_to_name(
    index: Union[Sequence[int], TensorLike, Sequence],
    idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
    *,
    is_number_fn: Callable[[Any], bool] = is_number_like,
) -> List[T_Name]:
    """Convert indices of labels to names using a mapping for **multiclass** classification.

    Args:
        indices: List of list of label indices.
        idx_to_name: Mapping to convert a class index to its name.
        is_number_fn: Type guard to check if a value is a scalar number. defaults to `is_number_like`.
    """
    if (
        isinstance(index, (Tensor, np.ndarray))
        and nelement(index) == 0
        and index.ndim > 1
    ):
        msg = f"Found 0 elements in {index=} but {index.ndim=} > 1, which means that we will lose information about shape when converting to names."
        warn_once(msg, __name__)

    def _impl(x) -> Union[T_Name, list]:
        if is_number_fn(x):
            return idx_to_name[to_item(x)]  # type: ignore
        elif isinstance(x, Iterable):
            return [_impl(xi) for xi in x]
        else:
            msg = f"Invalid argument {x=}. (not present in idx_to_name and not an iterable type)"
            raise ValueError(msg)

    name = _impl(index)
    return name  # type: ignore


def onehot_to_index(
    onehot: Tensor,
    *,
    padding_idx: Optional[int] = None,
    dim: int = -1,
) -> LongTensor:
    """Convert onehot boolean encoding to indices of labels for **multiclass** classification.

    Args:
        onehot: Onehot labels encoded as 2D matrix.
        padding_idx: Class index placeholder when input contains only zeroes. defaults to None.
        dim: Dimension of classes. defaults to -1.
    """
    onehot = onehot.int()
    index = onehot.argmax(dim=dim)

    if padding_idx is not None:
        empty = onehot.eq(False).all(dim=dim)
        index = torch.where(empty, padding_idx, index)

    return index  # type: ignore


def onehot_to_name(
    onehot: Tensor,
    idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
    *,
    dim: int = -1,
) -> List[T_Name]:
    """Convert onehot boolean encoding to names using a mapping for **multiclass** classification.

    Args:
        onehot: Onehot labels encoded as 2D matrix.
        idx_to_name: Mapping to convert a class index to its name.
        dim: Dimension of classes. defaults to -1.
    """
    indices = onehot_to_index(onehot, dim=dim)
    names = index_to_name(indices, idx_to_name)
    return names


def name_to_index(
    name: List[T_Name],
    idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
) -> Tensor:
    """Convert names to indices of labels for **multiclass** classification.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
    """
    if isinstance(idx_to_name, Mapping):
        name_to_idx = {name: idx for idx, name in idx_to_name.items()}
    else:
        name_to_idx = {name: idx for idx, name in enumerate(idx_to_name)}
    del idx_to_name

    def _impl(x) -> Union[T_Name, list]:
        if isinstance(x, Hashable) and x in name_to_idx:
            return name_to_idx[x]  # type: ignore
        elif isinstance(x, Iterable):
            return [_impl(xi) for xi in x]
        else:
            msg = f"Invalid argument {x=}. (not present in name_to_idx and not an iterable type)"
            raise ValueError(msg)

    index = _impl(name)
    index = torch.as_tensor(index, dtype=torch.long)
    return index  # type: ignore


def name_to_onehot(
    name: List[T_Name],
    idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
    *,
    device: DeviceLike = None,
    dtype: DTypeLike = torch.bool,
) -> Tensor:
    """Convert names to onehot boolean encoding for **multiclass** classification.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    index = name_to_index(name, idx_to_name)
    onehot = index_to_onehot(index, len(idx_to_name), device=device, dtype=dtype)
    return onehot


def probs_to_index(
    probs: Tensor,
    *,
    dim: int = -1,
) -> LongTensor:
    """Convert matrix of probabilities to indices of labels for **multiclass** classification.

    Args:
        probs: Output probabilities for each classes.
        dim: Dimension of classes. defaults to -1.
    """
    index = probs.argmax(dim=dim)
    return index  # type: ignore


def probs_to_onehot(
    probs: Tensor,
    *,
    dim: int = -1,
    device: DeviceLike = None,
    dtype: DTypeLike = torch.bool,
) -> Tensor:
    """Convert matrix of probabilities to onehot boolean encoding for **multiclass** classification.

    Args:
        probs: Output probabilities for each classes.
        dim: Dimension of classes. defaults to -1.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    if device is None:
        device = probs.device
    indices = probs_to_index(probs, dim=dim)
    onehot = index_to_onehot(indices, probs.shape[-1], device=device, dtype=dtype)
    return onehot


def probs_to_name(
    probs: Tensor,
    idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
    *,
    dim: int = -1,
) -> List[T_Name]:
    """Convert matrix of probabilities to labels names for **multiclass** classification.

    Args:
        probs: Output probabilities for each classes.
        idx_to_name: Mapping to convert a class index to its name.
        dim: Dimension of classes. defaults to -1.
    """
    indices = probs_to_index(probs, dim=dim)
    names = index_to_name(indices, idx_to_name)
    return names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import unittest
from typing import Any, List, Tuple, Type
from unittest import TestCase

import torch
from torch import Tensor, nn

from torchoutil.nn.modules import tensor as tensor_module


def module_name_to_fn_name(x: str) -> str:
    if x.isupper() or x.islower():
        return x.lower()

    x = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", x)
    x = x.lower()
    return x


class TestModuleTensorCompat(TestCase):
    def test_all_results(self) -> None:
        base_modules = [torch, torch.Tensor, torch.nn.functional, torch.fft]
        all_fn_names = {
            base_module: dict.fromkeys(
                name
                for name in dir(base_module)
                if callable(getattr(base_module, name))
            )
            for base_module in base_modules
        }

        tested_modules = set()
        missed_modules = set()

        num_modules = 0
        targets: List[Tuple[Type[nn.Module], Any, str]] = []

        for name in dir(tensor_module):
            module_cls = getattr(tensor_module, name)
            if (
                not isinstance(module_cls, type)
                or not issubclass(module_cls, nn.Module)
                or module_cls is nn.Module
            ):
                continue

            num_modules += 1
            fn_name = module_name_to_fn_name(name)

            module_targets: List[Tuple[Type[nn.Module], Any, str]] = []
            for base_module, fn_names in all_fn_names.items():
                if fn_name not in fn_names:
                    continue
                target = module_cls, base_module, fn_name
                module_targets.append(target)

            if len(module_targets) == 0:
                missed_modules.add(name)
            else:
                tested_modules.add(name)

            targets += module_targets

        x = torch.rand((10,))
        for module_cls, base_module, fn_name in targets:
            try:
                module = module_cls()
                result = module(x)
            except Exception as err:
                result = err

            try:
                fn = getattr(base_module, fn_name)
                expected = fn(x)
            except Exception as err:
                expected = err

            if isinstance(result, Exception) and isinstance(expected, Exception):
                pass
            elif isinstance(result, Tensor) and isinstance(expected, Tensor):
                assert torch.equal(result, expected)
            else:
                assert (
                    result == expected
                ), f"{result=} ; {expected=} from {module_cls.__qualname__} and {base_module.__name__}.{fn_name}"

        print(f"Total base coverage hit: {len(tested_modules)}")
        print(f"Hit_: {tested_modules}")
        print(f"Miss: {missed_modules}")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Helper functions for conversion between classes indices, multihot, multi-names and probabilities for multilabel classification.
"""

from typing import Hashable, Iterable, List, Mapping, Optional, Sequence, TypeVar, Union

import torch
from torch import Tensor

from torchoutil.core.get import DeviceLike, DTypeLike, get_device, get_dtype
from torchoutil.nn.functional.others import can_be_stacked, to_item
from torchoutil.nn.functional.pad import pad_and_stack_rec
from torchoutil.nn.functional.transform import to_tensor
from torchoutil.pyoutil.typing import is_sequence_int
from torchoutil.types import LongTensor, is_number_like, is_tensor_like
from torchoutil.types._typing import TensorLike

T_Name = TypeVar("T_Name", bound=Hashable)


def indices_to_multihot(
    indices: Union[Sequence[Union[Sequence[int], TensorLike]], TensorLike],
    num_classes: int,
    *,
    padding_idx: Optional[int] = None,
    device: DeviceLike = None,
    dtype: DTypeLike = torch.bool,
) -> Tensor:
    """Convert indices of labels to multihot boolean encoding for **multilabel** classification.

    Args:
        indices: List of list of label indices. Values should be integers in range [0..num_classes-1]
        num_classes: Number maximal of unique classes.
        padding_idx: Optional pad value to ignore.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    if device is None and isinstance(indices, Tensor):
        device = indices.device
    else:
        device = get_device(device)
    dtype = get_dtype(dtype)

    def _impl(x) -> Tensor:
        if is_tensor_like(x) and not _is_valid_indices(x):
            raise ValueError(f"Invalid argument shape {x.shape=}.")

        if (is_tensor_like(x) and x.ndim == 1) or is_sequence_int(x):
            x = torch.as_tensor(x, dtype=torch.long, device=device)

            if padding_idx is not None:
                x = torch.where(x == padding_idx, num_classes, x)
                target_num_classes = num_classes + 1
            else:
                target_num_classes = num_classes

            multihot = torch.full(
                (target_num_classes,),
                False,
                dtype=dtype,
                device=device,
            )
            multihot.scatter_(0, x, True)

            if padding_idx is not None:
                multihot = multihot[..., :-1]

            return multihot

        elif isinstance(x, Iterable):
            result = [_impl(xi) for xi in x]
            return torch.stack(result)

        else:
            raise ValueError(f"Invalid argument {x=}.")

    multihot = _impl(indices)
    return multihot


def indices_to_multinames(
    indices: Union[Sequence[Union[Sequence[int], TensorLike]], TensorLike],
    idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
    *,
    padding_idx: Optional[int] = None,
) -> List[List[T_Name]]:
    """Convert indices of labels to names using a mapping for **multilabel** classification.

    Args:
        indices: List of list of label indices.
        idx_to_name: Mapping to convert a class index to its name.
        padding_idx: Optional pad value to ignore.
    """
    if not isinstance(indices, Iterable):
        raise TypeError(f"Invalid argument {indices=}. (not an iterable)")

    def _impl(x) -> Union[int, list]:
        if is_number_like(x):
            return idx_to_name[to_item(x)]  # type: ignore
        elif isinstance(x, Iterable):
            return [
                _impl(xi)
                for xi in x
                if padding_idx is None or not is_number_like(xi) or xi != padding_idx
            ]
        else:
            msg = f"Invalid argument {x=}. (not present in idx_to_name and not an iterable type)"
            raise ValueError(msg)

    names = _impl(indices)
    return names  # type: ignore


def multihot_to_indices(
    multihot: Union[
        TensorLike, Sequence[TensorLike], Sequence[Sequence[bool]], Sequence
    ],
    *,
    keep_tensor: bool = False,
    padding_idx: Optional[int] = None,
    dim: int = -1,
) -> Union[List, LongTensor]:
    """Convert multihot boolean encoding to indices of labels for **multilabel** classification.

    Args:
        multihot: Multihot labels encoded as 2D matrix. Must be convertible to Tensor.
        keep_tensor: If True, output will be converted to a tensor if possible. defaults to False.
        padding_idx: Class index fill value. When none, output will not be padded. defaults to None.
        dim: Dimension of classes. defaults to -1.
    """
    multihot = to_tensor(multihot)
    multihot = multihot.transpose(dim, -1)

    if not _is_valid_indices(multihot):
        msg = f"Invalid argument shape {multihot=}. (expected first axis should be > 0)"
        raise ValueError(msg)

    def _impl(x: TensorLike) -> Union[list, LongTensor]:
        if x.ndim == 1:
            x = torch.as_tensor(x, dtype=torch.bool)
            preds = torch.where(x)[0]
            if not keep_tensor:
                preds = preds.tolist()
            return preds  # type: ignore

        elif x.ndim > 1:
            preds = [_impl(multihot_i) for multihot_i in x]
            if padding_idx is not None:
                preds = pad_and_stack_rec(preds, padding_idx, dtype=torch.long)
                if not keep_tensor:
                    preds = preds.tolist()
            elif keep_tensor and can_be_stacked(preds):
                preds = torch.stack(preds)
            return preds  # type: ignore

        else:
            msg = f"Invalid argument {x=}. (found {x.ndim} dims)"
            raise ValueError(msg)

    result = _impl(multihot)
    return result


def multihot_to_multinames(
    multihot: Union[TensorLike, Sequence[TensorLike], Sequence[Sequence[bool]]],
    idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
    *,
    dim: int = -1,
) -> List[List[T_Name]]:
    """Convert multihot boolean encoding to names using a mapping for **multilabel** classification.

    Args:
        multihot: Multihot labels encoded as 2D matrix.
        idx_to_name: Mapping to convert a class index to its name.
        dim: Dimension of classes. defaults to -1.
    """
    indices = multihot_to_indices(multihot, dim=dim)
    names = indices_to_multinames(indices, idx_to_name)
    return names


def multinames_to_indices(
    names: List[List[T_Name]],
    idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
) -> List[List[int]]:
    """Convert names to indices of labels for **multilabel** classification.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
    """
    if isinstance(idx_to_name, Mapping):
        name_to_idx = {name: idx for idx, name in idx_to_name.items()}
    else:
        name_to_idx = {name: idx for idx, name in enumerate(idx_to_name)}
    del idx_to_name

    def _impl(x) -> Union[int, list]:
        if isinstance(x, Hashable) and x in name_to_idx:
            return name_to_idx[x]  # type: ignore
        elif isinstance(x, Iterable):
            return [_impl(xi) for xi in x]
        else:
            msg = f"Invalid argument {x=}. (not present in idx_to_name and not an iterable type)"
            raise ValueError(msg)

    indices = _impl(names)
    return indices  # type: ignore


def multinames_to_multihot(
    names: List[List[T_Name]],
    idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
    *,
    device: DeviceLike = None,
    dtype: DTypeLike = torch.bool,
) -> Tensor:
    """Convert names to multihot boolean encoding for **multilabel** classification.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    indices = multinames_to_indices(names, idx_to_name)
    multihot = indices_to_multihot(
        indices,
        len(idx_to_name),
        device=device,
        dtype=dtype,
    )
    return multihot


def probs_to_indices(
    probs: TensorLike,
    threshold: Union[float, Sequence[float], TensorLike],
    *,
    padding_idx: Optional[int] = None,
    dim: int = -1,
) -> Union[List, LongTensor]:
    """Convert matrix of probabilities to indices of labels for **multilabel** classification.

    Args:
        probs: Output probabilities for each classes.
        threshold: Threshold(s) to binarize probabilities. Can be a scalar or a sequence of (num_classes,) thresholds.
        padding_idx: Class index fill value. When none, output will not be padded. defaults to None.
        dim: Dimension of classes. defaults to -1.
    """
    multihot = probs_to_multihot(probs, threshold, dim=dim)
    indices = multihot_to_indices(multihot, padding_idx=padding_idx, dim=dim)
    return indices


def probs_to_multihot(
    probs: TensorLike,
    threshold: Union[float, Sequence[float], TensorLike],
    *,
    dim: int = -1,
    device: DeviceLike = None,
    dtype: DTypeLike = torch.bool,
) -> Tensor:
    """Convert matrix of probabilities to multihot boolean encoding for **multilabel** classification.

    Args:
        probs: Output probabilities for each class.
        threshold: Threshold(s) to binarize probabilities. Can be a scalar or a sequence of (num_classes,) thresholds.
        dim: Dimension of classes. defaults to -1.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    probs = to_tensor(probs)
    if probs.ndim == 0:
        msg = f"Invalid argument ndim {probs.ndim=}. (expected at least 1 dim)."
        raise ValueError(msg)

    num_classes = probs.shape[dim]
    if device is None:
        device = probs.device
    else:
        device = get_device(device)
    dtype = get_dtype(dtype)

    if not isinstance(threshold, Tensor):
        threshold = torch.as_tensor(threshold, dtype=probs.dtype, device=device)
    else:
        threshold = threshold.to(device=device)

    if threshold.ndim == 0:
        threshold = threshold.repeat(num_classes)
    elif threshold.ndim == 1:
        if threshold.shape[0] != num_classes:
            msg = f"Invalid argument threshold. (number of thresholds is {threshold.shape[0]} but found {num_classes} classes)"
            raise ValueError(msg)
    else:
        msg = f"Invalid number of dimensions in input threshold. (found {threshold.shape=} but expected 0-d or 1-d tensor)"
        raise ValueError(msg)

    dim = dim % probs.ndim
    slices = [(slice(None) if i == dim else None) for i in range(probs.ndim)]
    threshold = threshold[slices]

    multihot = probs >= threshold
    multihot = multihot.to(dtype=dtype)
    return multihot


def probs_to_multinames(
    probs: TensorLike,
    threshold: Union[float, Sequence[float], TensorLike],
    idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
) -> List[List[T_Name]]:
    """Convert matrix of probabilities to labels names for **multilabel** classification.

    Args:
        probs: Output probabilities for each classes.
        threshold: Threshold(s) to binarize probabilities. Can be a scalar or a sequence of (num_classes,) thresholds.
        idx_to_name: Mapping to convert a class index to its name.
    """
    indices = probs_to_indices(probs, threshold)
    names = indices_to_multinames(indices, idx_to_name)
    return names


def _is_valid_indices(x: TensorLike, dim: int = -1) -> bool:
    """Returns True if tensor has valid shape for indices_to_multihot."""
    shape = list(x.shape)
    dim = dim % len(shape)
    try:
        idx = shape.index(0)
        return idx == dim
    except ValueError:
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import functools
import itertools
import math
import pickle
import struct
import zlib
from dataclasses import asdict
from types import FunctionType, MethodType
from typing import Any, Callable, Iterable, Literal, Mapping, Union

import torch
from torch import Tensor, nn

from torchoutil.core.packaging import _NUMPY_AVAILABLE
from torchoutil.extras.numpy import np
from torchoutil.nn.functional.others import is_complex, is_floating_point, nelement
from torchoutil.pyoutil.inspect import get_fullname
from torchoutil.pyoutil.typing import (
    BuiltinNumber,
    BuiltinScalar,
    DataclassInstance,
    NamedTupleInstance,
    NoneType,
)

Checksumable = Union[
    int,
    bool,
    complex,
    float,
    NoneType,
    str,
    bytes,
    bytearray,
    nn.Module,
    Tensor,
    np.ndarray,
    np.generic,
    NamedTupleInstance,
    DataclassInstance,
    Mapping,
    Iterable,
    MethodType,
    FunctionType,
    functools.partial,
    type,
]
CHECKSUMABLE_TYPES = (
    "int",
    "bool",
    "complex",
    "float",
    "NoneType",
    "str",
    "bytes",
    "bytearray",
    "torch.nn.Module",
    "torch.Tensor",
    "numpy.ndarray",
    "numpy.generic",
    "NamedTuple",
    "Dataclass",
    "Mapping",
    "Iterable",
    "MethodType",
    "FunctionType",
    "functools.partial",
    "type",
)
UnkMode = Literal["pickle", "error"]
UNK_MODES = ("pickle", "error")


# Recursive functions
def checksum(
    x: Checksumable,
    *,
    unk_mode: UnkMode = "error",
    allow_protocol: bool = True,
    **kwargs,
) -> int:
    """Alias for `torchoutil.checksum_any`."""
    return checksum_any(
        x,
        unk_mode=unk_mode,
        allow_protocol=allow_protocol,
        **kwargs,
    )


def checksum_any(
    x: Checksumable,
    *,
    unk_mode: UnkMode = "error",
    allow_protocol: bool = True,
    **kwargs,
) -> int:
    """Compute checksum of an arbitrary python object."""
    kwargs.update(
        dict(
            unk_mode=unk_mode,
            allow_protocol=allow_protocol,
        )
    )
    if isinstance(x, (int, bool, complex, float)):
        return checksum_builtin_number(x, **kwargs)
    elif x is None:
        return checksum_none(x, **kwargs)
    elif isinstance(x, str):
        return checksum_str(x, **kwargs)
    elif isinstance(x, bytes):
        return checksum_bytes(x, **kwargs)
    elif isinstance(x, bytearray):
        return checksum_bytearray(x, **kwargs)
    elif isinstance(x, nn.Module):
        return checksum_module(x, **kwargs)
    elif isinstance(x, Tensor):
        return checksum_tensor(x, **kwargs)
    elif isinstance(x, (np.ndarray, np.generic)):
        return checksum_ndarray(x, **kwargs)
    elif allow_protocol and isinstance(x, NamedTupleInstance):
        return checksum_namedtuple(x, **kwargs)
    elif allow_protocol and isinstance(x, DataclassInstance):
        return checksum_dataclass(x, **kwargs)
    elif allow_protocol and isinstance(x, Mapping):
        return checksum_mapping(x, **kwargs)
    elif allow_protocol and isinstance(x, Iterable):
        return checksum_iterable(x, **kwargs)
    elif isinstance(x, MethodType):
        return checksum_method(x, **kwargs)
    elif isinstance(x, FunctionType):
        return checksum_function(x, **kwargs)
    elif isinstance(x, functools.partial):
        return checksum_partial(x, **kwargs)
    elif isinstance(x, type):
        return checksum_type(x, **kwargs)
    elif unk_mode == "pickle":
        return checksum_bytes(pickle.dumps(x), **kwargs)
    elif unk_mode == "error":
        msg = f"Invalid argument type {type(x)}. (expected one of {CHECKSUMABLE_TYPES})"
        raise TypeError(msg)
    else:
        raise ValueError(f"Invalid argument {unk_mode=}. (expected one of {UNK_MODES})")


def checksum_dataclass(x: DataclassInstance, **kwargs) -> int:
    accumulator = kwargs.pop("accumulator", 0) + checksum_str(get_fullname(x), **kwargs)
    kwargs["accumulator"] = accumulator
    return checksum_mapping(asdict(x), **kwargs)


def checksum_iterable(x: Iterable[Any], **kwargs) -> int:
    accumulator = kwargs.pop("accumulator", 0) + checksum_str(get_fullname(x), **kwargs)
    csum = sum(
        checksum_any(xi, accumulator=accumulator + (i + 1), **kwargs) * (i + 1)
        for i, xi in enumerate(x)
    )
    return csum + accumulator


def checksum_mapping(x: Mapping[Any, Any], **kwargs) -> int:
    return checksum_iterable(x.items(), **kwargs)


def checksum_method(x: MethodType, **kwargs) -> int:
    fn = getattr(x.__self__, x.__name__)
    checksums = [
        checksum_any(x.__self__, **kwargs),
        checksum_function(fn, **kwargs),
    ]
    return checksum_iterable(checksums, **kwargs)


def checksum_module(
    x: nn.Module,
    *,
    only_trainable: bool = False,
    with_names: bool = False,
    buffers: bool = False,
    training: bool = False,
    **kwargs,
) -> int:
    """Compute a simple checksum over module parameters."""
    training = x.training
    x.train(training)

    if with_names:
        params_it = (
            (n, p)
            for n, p in x.named_parameters()
            if not only_trainable or p.requires_grad
        )
    else:
        params_it = (
            param
            for param in x.parameters()
            if not only_trainable or param.requires_grad
        )

    if not buffers:
        iterator = params_it
    elif with_names:
        buffers_it = (name_buffer for name_buffer in x.named_buffers())
        iterator = itertools.chain(params_it, buffers_it)
    else:
        buffers_it = (buffer for buffer in x.buffers())
        iterator = itertools.chain(params_it, buffers_it)

    csum = checksum_iterable(iterator, **kwargs)
    x.train(training)
    return csum


def checksum_namedtuple(x: NamedTupleInstance, **kwargs) -> int:
    accumulator = kwargs.pop("accumulator", 0) + checksum_str(get_fullname(x), **kwargs)
    kwargs["accumulator"] = accumulator
    return checksum_mapping(x._asdict(), **kwargs)


def checksum_partial(x: functools.partial, **kwargs) -> int:
    return checksum_iterable(("functools.partial", x.func, x.args, x.keywords))


# Intermediate functions
def checksum_builtin_number(x: BuiltinNumber, **kwargs) -> int:
    """Compute a simple checksum of a builtin scalar number."""
    # Note: instance check must follow this order: bool, int, float, complex, because isinstance(True, int) returns True !
    if isinstance(x, bool):
        return checksum_bool(x, **kwargs)
    elif isinstance(x, int):
        return checksum_int(x, **kwargs)
    elif isinstance(x, float):
        return checksum_float(x, **kwargs)
    elif isinstance(x, complex):
        return checksum_complex(x, **kwargs)
    else:
        BUILTIN_NUMBER_TYPES = ("bool", "int", "float", "complex")
        msg = (
            f"Invalid argument type {type(x)}. (expected one of {BUILTIN_NUMBER_TYPES})"
        )
        raise TypeError(msg)


def checksum_builtin_scalar(x: BuiltinScalar, **kwargs) -> int:
    if isinstance(x, BuiltinNumber):
        return checksum_builtin_number(x, **kwargs)
    elif isinstance(x, bytes):
        return checksum_bytes(x, **kwargs)
    elif isinstance(x, NoneType):
        return checksum_none(x, **kwargs)
    elif isinstance(x, str):
        return checksum_str(x, **kwargs)
    else:
        msg = f"Invalid argument type {type(x)}. (expected int, bool, complex float, bytes, None, or str)"
        raise TypeError(msg)


def checksum_bytearray(x: bytearray, **kwargs) -> int:
    accumulator = kwargs.pop("accumulator", 0) + checksum_str(get_fullname(x), **kwargs)
    return _checksum_bytes_bytearray(x, accumulator=accumulator, **kwargs)


def checksum_complex(x: complex, **kwargs) -> int:
    return checksum_tensor(torch.as_tensor([x.real, x.imag]), **kwargs)


def checksum_function(x: FunctionType, **kwargs) -> int:
    return checksum_str(x.__qualname__, **kwargs)


def checksum_none(x: None, **kwargs) -> int:
    return checksum_type(x.__class__, **kwargs) + kwargs.get("accumulator", 0)


def checksum_ndarray(x: Union[np.ndarray, np.generic], **kwargs) -> int:
    if not _NUMPY_AVAILABLE:
        msg = "Cannot call function 'checksum_ndarray' because optional dependancy 'numpy' is not installed. Please install it using 'pip install torchoutil[extras]'"
        raise NotImplementedError(msg)

    # Supports non-numeric numpy arrays (byte string, unicode string, object, void)
    if x.dtype.kind in ("S", "U", "O", "V"):
        return checksum_any(x.tolist(), **kwargs)

    return _checksum_tensor_array_like(
        x,
        nan_to_num_fn=np.nan_to_num,
        arange_fn=np.arange,
        **kwargs,
    )


def checksum_str(x: str, **kwargs) -> int:
    return checksum_bytes(x.encode(), **kwargs)


@torch.inference_mode()
def checksum_tensor(x: Tensor, **kwargs) -> int:
    """Compute a simple checksum of a tensor. Order of values matter for the checksum."""
    return _checksum_tensor_array_like(
        x,
        nan_to_num_fn=torch.nan_to_num,
        arange_fn=torch.arange,
        **kwargs,
    )


def checksum_type(x: type, **kwargs) -> int:
    return checksum_str(x.__qualname__, **kwargs)


# Terminate functions
def checksum_bool(x: bool, **kwargs) -> int:
    return int(x) + kwargs.get("accumulator", 0)


def checksum_bytes(x: Union[bytes, bytearray], **kwargs) -> int:
    return _checksum_bytes_bytearray(x, **kwargs)


def checksum_float(x: float, **kwargs) -> int:
    xint = _interpret_float_as_int(x)
    return xint + kwargs.get("accumulator", 0)


def checksum_int(x: int, **kwargs) -> int:
    return x + kwargs.get("accumulator", 0)


def _interpret_float_as_int(x: float) -> int:
    xbytes = struct.pack("!f", x)
    xint = struct.unpack("!i", xbytes)[0]
    return xint


def _checksum_bytes_bytearray(x: Union[bytes, bytearray], **kwargs) -> int:
    return zlib.crc32(x) % (1 << 32) + kwargs.get("accumulator", 0)


def _checksum_tensor_array_like(
    x: Union[Tensor, np.ndarray, np.generic],
    *,
    nan_to_num_fn: Callable,
    arange_fn: Callable,
    **kwargs,
) -> int:
    if x.ndim == 0:
        return checksum_builtin_number(x.item(), **kwargs)

    if x.dtype == bool:
        range_dtype = int
    elif x.dtype == torch.bool:
        range_dtype = torch.int
    elif is_complex(x):
        range_dtype = x.real.dtype  # type: ignore
    else:
        range_dtype = x.dtype

    if is_floating_point(x) or is_complex(x):
        nan_csum = checksum_float(math.nan, **kwargs)
        neginf_csum = checksum_float(-math.inf, **kwargs)
        posinf_csum = checksum_float(math.inf, **kwargs)
        x = nan_to_num_fn(
            x,
            nan=nan_csum,
            neginf=neginf_csum,
            posinf=posinf_csum,
        )

    shape = x.shape  # type: ignore
    x = x.flatten()  # type: ignore
    arange = arange_fn(1, nelement(x) + 1, dtype=range_dtype)
    x = x + arange
    x = x * arange
    x = x.sum()

    # Ensure that accumulator exists
    kwargs["accumulator"] = kwargs.get("accumulator", 0)
    type_csum = checksum_str(get_fullname(x), **kwargs)

    kwargs["accumulator"] += type_csum
    shape_csum = checksum_iterable(shape, **kwargs)

    kwargs["accumulator"] += shape_csum
    x_csum = checksum_builtin_number(x.item(), **kwargs)

    return x_csum



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Module versions of tensor functions that do not already exists in PyTorch."""

from typing import Any, List, Optional, Sequence, Tuple, Union, overload

import torch
from torch import Tensor, nn
from torch.nn import functional as F
from torch.types import Device, Number

from torchoutil.pyoutil.collections import dump_dict
from torchoutil.types import BuiltinNumber
from torchoutil.utils import return_types


class Abs(nn.Module):
    """
    Module version of :func:`~torch.abs`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.abs()


class Angle(nn.Module):
    """
    Module version of :func:`~torch.angle`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.angle()


class AsTensor(nn.Module):
    """
    Module version of :func:`~torch.as_tensor`.
    """

    def __init__(
        self,
        *,
        device: Device = None,
        dtype: Optional[torch.dtype] = None,
    ) -> None:
        super().__init__()
        self.device = device
        self.dtype = dtype

    def forward(self, x: Any) -> Tensor:
        return torch.as_tensor(x, dtype=self.dtype, device=self.device)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dtype=self.dtype,
                device=self.device,
            ),
            ignore_lst=(None,),
        )


class Exp(nn.Module):
    """
    Module version of :func:`~torch.exp`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.exp()


class Exp2(nn.Module):
    """
    Module version of :func:`~torch.exp2`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.exp2()


class FFT(nn.Module):
    """
    Module version of :func:`~torch.fft.fft`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return torch.fft.fft(x)


class IFFT(nn.Module):
    """
    Module version of :func:`~torch.fft.ifft`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return torch.fft.ifft(x)


class Imag(nn.Module):
    """
    Module version of :func:`~torch.Tensor.imag`.
    """

    def __init__(self, *, return_zeros: bool = False) -> None:
        """Return the imaginary part of a complex tensor.

        Args:
            return_zeros:
                If True and the input is not a complex tensor, the module will return a tensor of same shape containing zeros.
                If False and the input is not a complex tensor, raises the default RuntimError of PyTorch.
        """
        super().__init__()
        self.return_zeros = return_zeros

    def forward(self, x: Tensor) -> Tensor:
        if self.return_zeros and not x.is_complex():
            return torch.zeros_like(x)
        else:
            return x.imag


class Interpolate(nn.Module):
    """
    Module version of :func:`~torch.nn.functional.interpolate`.
    """

    def __init__(
        self,
        size: Union[int, Tuple[int, ...], None] = None,
        scale_factor: Union[float, Tuple[float, ...], None] = None,
        mode: str = "nearest",
        align_corners: Optional[bool] = None,
        recompute_scale_factor: Optional[bool] = None,
        antialias: bool = False,
    ) -> None:
        super().__init__()
        self.size = size
        self.scale_factor = scale_factor
        self.mode = mode
        self.align_corners = align_corners
        self.recompute_scale_factor = recompute_scale_factor
        self.antialias = antialias

    def forward(self, x: Tensor) -> Tensor:
        return F.interpolate(
            x,
            size=self.size,
            scale_factor=self.scale_factor,
            mode=self.mode,
            align_corners=self.align_corners,
            recompute_scale_factor=self.recompute_scale_factor,
            antialias=self.antialias,
        )


class Log(nn.Module):
    """
    Module version of :func:`~torch.log`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.log()


class Log10(nn.Module):
    """
    Module version of :func:`~torch.log10`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.log10()


class Log2(nn.Module):
    """
    Module version of :func:`~torch.log2`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.log2()


class Max(nn.Module):
    """
    Module version of :func:`~torch.max`.
    """

    def __init__(
        self,
        dim: Optional[int] = None,
        return_values: bool = True,
        return_indices: Optional[bool] = None,
        keepdim: bool = False,
    ) -> None:
        if return_indices is None:
            return_indices = dim is not None
        if not return_values and not return_indices:
            msg = f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)
        if dim is None and keepdim:
            msg = f"Invalid combinaison of arguments {dim=} and {keepdim=}. (expected dim is not None or keepdim=False)"
            raise ValueError(msg)

        super().__init__()
        self.dim = dim
        self.return_values = return_values
        self.return_indices = return_indices
        self.keepdim = keepdim

    def forward(self, x: Tensor) -> Union[Tensor, return_types.max]:
        if self.dim is None:
            index = x.argmax()
            values_indices = return_types.max([x.flatten()[index], index])
        else:
            values_indices = x.max(dim=self.dim, keepdim=self.keepdim)

        if self.return_values and self.return_indices:
            return values_indices  # type: ignore
        elif self.return_values:
            return values_indices.values
        elif self.return_indices:
            return values_indices.indices
        else:
            msg = f"Invalid combinaison of arguments {self.return_values=} and {self.return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                return_values=self.return_values,
                return_indices=self.return_indices,
                keepdim=self.keepdim,
            ),
        )


class Mean(nn.Module):
    """
    Module version of :func:`~torch.mean`.
    """

    def __init__(self, dim: Optional[int] = None, keepdim: bool = False) -> None:
        super().__init__()
        self.dim = dim
        self.keepdim = keepdim

    def forward(self, x: Tensor) -> Tensor:
        if self.dim is None:
            return x.mean()
        else:
            return x.mean(dim=self.dim, keepdim=self.keepdim)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                keepdim=self.keepdim,
            ),
            ignore_lst=(None,),
        )


class Min(nn.Module):
    """
    Module version of :func:`~torch.min`.
    """

    def __init__(
        self,
        dim: Optional[int] = None,
        return_values: bool = True,
        return_indices: Optional[bool] = None,
        keepdim: bool = False,
    ) -> None:
        if return_indices is None:
            return_indices = dim is not None
        if not return_values and not return_indices:
            msg = f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)
        if dim is None and keepdim:
            msg = f"Invalid combinaison of arguments {dim=} and {keepdim=}. (expected dim is not None or keepdim=False)"
            raise ValueError(msg)

        super().__init__()
        self.dim = dim
        self.return_values = return_values
        self.return_indices = return_indices
        self.keepdim = keepdim

    def forward(self, x: Tensor) -> Union[Tensor, return_types.min]:
        if self.dim is None:
            index = x.argmin()
            values_indices = return_types.min([x.flatten()[index], index])
        else:
            values_indices = x.min(dim=self.dim, keepdim=self.keepdim)

        if self.return_values and self.return_indices:
            return values_indices  # type: ignore
        elif self.return_values:
            return values_indices.values
        elif self.return_indices:
            return values_indices.indices
        else:
            msg = f"Invalid combinaison of arguments {self.return_values=} and {self.return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                return_values=self.return_values,
                return_indices=self.return_indices,
                keepdim=self.keepdim,
            ),
        )


class Normalize(nn.Module):
    """
    Module version of :func:`~torch.nn.functional.normalize`.
    """

    def __init__(
        self,
        p: float = 2.0,
        dim: int = 1,
        eps: float = 1e-12,
    ) -> None:
        super().__init__()
        self.p = p
        self.dim = dim
        self.eps = eps

    def forward(self, x: Tensor) -> Tensor:
        return F.normalize(x, self.p, self.dim, self.eps)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                p=self.p,
                dim=self.dim,
                eps=self.eps,
            )
        )


class Permute(nn.Module):
    """
    Module version of :func:`~torch.permute`.
    """

    def __init__(self, *args: int) -> None:
        super().__init__()
        self.dims = tuple(args)

    def forward(self, x: Tensor) -> Tensor:
        return x.permute(self.dims)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dims=self.dims,
            ),
            fmt="{value}",
        )


class Pow(nn.Module):
    """
    Module version of :func:`~torch.Tensor.pow`.
    """

    def __init__(self, exponent: Union[Number, Tensor]) -> None:
        super().__init__()
        self.exponent = exponent

    def forward(self, x: Tensor) -> Tensor:
        return x.pow(self.exponent)

    def extra_repr(self) -> str:
        return dump_dict(dict(exponent=self.exponent))


class Real(nn.Module):
    """
    Module version of :func:`~torch.Tensor.real`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.real


class Repeat(nn.Module):
    """
    Module version of :func:`~torch.repeat`.
    """

    def __init__(self, *repeats: int) -> None:
        super().__init__()
        self.repeats = repeats

    def forward(self, x: Tensor) -> Tensor:
        return x.repeat(self.repeats)

    def extra_repr(self) -> str:
        return dump_dict(dict(repeats=self.repeats))


class RepeatInterleave(nn.Module):
    """
    Module version of :func:`~torch.repeat_interleave`.
    """

    def __init__(
        self,
        repeats: Union[int, Tensor],
        dim: int,
        output_size: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.repeats = repeats
        self.dim = dim
        self.output_size = output_size

    def forward(self, x: Tensor) -> Tensor:
        return x.repeat_interleave(self.repeats, self.dim, output_size=self.output_size)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                repeats=self.repeats,
                dim=self.dim,
                output_size=self.output_size,
            ),
            ignore_lst=(None,),
        )


class Reshape(nn.Module):
    """
    Module version of :func:`~torch.reshape`.
    """

    def __init__(self, *shape: int) -> None:
        super().__init__()
        self.shape = shape

    def forward(self, x: Tensor) -> Tensor:
        return x.reshape(self.shape)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                shape=self.shape,
            ),
        )


class Squeeze(nn.Module):
    """
    Module version of :func:`~torch.squeeze`.
    """

    def __init__(self, dim: Optional[int] = None, inplace: bool = True) -> None:
        super().__init__()
        self.dim = dim
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if not self.inplace:
            if self.dim is None:
                return x.squeeze()
            else:
                return x.squeeze(self.dim)
        else:
            if self.dim is None:
                return x.squeeze_()
            else:
                return x.squeeze_(self.dim)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                inplace=self.inplace,
            ),
        )


class TensorTo(nn.Module):
    """
    Module version of :func:`~torch.Tensor.to`.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()
        self.kwargs = kwargs

    def forward(self, x: Tensor) -> Tensor:
        return x.to(**self.kwargs)

    def extra_repr(self) -> str:
        return dump_dict(dict(self.kwargs))


class ToItem(nn.Module):
    """
    Module version of :func:`~torch.Tensor.item`.
    """

    def forward(self, x: Tensor) -> BuiltinNumber:
        return x.item()


class ToList(nn.Module):
    """
    Module version of :func:`~torch.Tensor.tolist`.
    """

    def forward(self, x: Tensor) -> List:
        return x.tolist()


class Topk(nn.Module):
    """
    Module version of :func:`~torch.Tensor.topk`.
    """

    def __init__(
        self,
        k: int,
        dim: int = -1,
        largest: bool = True,
        sorted: bool = True,
        return_values: bool = True,
        return_indices: bool = True,
    ) -> None:
        if not return_values and not return_indices:
            msg = f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)

        super().__init__()
        self.k = k
        self.dim = dim
        self.largest = largest
        self.sorted = sorted
        self.return_values = return_values
        self.return_indices = return_indices

    def forward(self, x: Tensor) -> Union[Tensor, return_types.topk]:
        values_indices = x.topk(
            k=self.k,
            dim=self.dim,
            largest=self.largest,
            sorted=self.sorted,
        )
        if self.return_values and self.return_indices:
            return values_indices
        elif self.return_values:
            return values_indices.values
        elif self.return_indices:
            return values_indices.indices
        else:
            msg = f"Invalid combinaison of arguments {self.return_values=} and {self.return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                k=self.k,
                dim=self.dim,
                largest=self.largest,
                sorted=self.sorted,
                return_values=self.return_values,
                return_indices=self.return_indices,
            ),
        )


class Transpose(nn.Module):
    """
    Module version of :func:`~torch.transpose`.
    """

    def __init__(self, dim0: int, dim1: int, copy: bool = False) -> None:
        super().__init__()
        self.dim0 = dim0
        self.dim1 = dim1
        self.copy = copy

    def forward(self, x: Tensor) -> Tensor:
        if self.copy:
            return torch.transpose_copy(x, self.dim0, self.dim1)
        else:
            return torch.transpose(x, self.dim0, self.dim1)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim0=self.dim0,
                dim1=self.dim1,
            ),
            fmt="{value}",
        )


class Unsqueeze(nn.Module):
    """
    Module version of :func:`~torch.Tensor.unsqueeze`.
    """

    def __init__(self, dim: int, inplace: bool = True) -> None:
        super().__init__()
        self.dim = dim
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if not self.inplace:
            return x.unsqueeze(self.dim)
        else:
            return x.unsqueeze_(self.dim)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                inplace=self.inplace,
            ),
        )


class View(nn.Module):
    @overload
    def __init__(self, dtype: torch.dtype) -> None:
        ...

    @overload
    def __init__(self, size: Sequence[int]) -> None:
        ...

    @overload
    def __init__(self, *size: int) -> None:
        ...

    def __init__(self, *args) -> None:
        super().__init__()
        self.args = args

    def forward(self, x: Tensor) -> Tensor:
        return x.view(*self.args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Dict, List, Literal, TypedDict

import h5py
import numpy as np

# Force this encoding
HDF_ENCODING = "utf-8"
# Key suffix to store tensor shapes (because they are padded in hdf file)
SHAPE_SUFFIX = "__shape"
# Type for strings
HDF_STRING_DTYPE: np.dtype = h5py.string_dtype(HDF_ENCODING, None)
# Type for empty lists
HDF_VOID_DTYPE: np.dtype = h5py.opaque_dtype("V1")


HDFItemType = Literal["dict", "tuple"]


class HDFDatasetAttributes(TypedDict):
    added_columns: List[str]
    creation_date: str
    encoding: str
    file_kwds: Dict[str, Any]
    global_hash_value: int
    info: Dict[str, Any]
    item_type: HDFItemType
    length: int
    load_as_complex: Dict[str, bool]
    pre_transform: str
    shape_suffix: str
    source_dataset: str
    src_np_dtypes: Dict[str, np.dtype]
    use_vlen_str: bool
    user_attrs: Any
    torchoutil_version: str


_DUMPED_JSON_KEYS = (
    "file_kwds",
    "info",
    "load_as_complex",
    "src_np_dtypes",
    "user_attrs",
)
_DEFAULTS_RAW_HDF_ATTRIBUTES = {
    "added_columns": [],
    "creation_date": "unknown",
    "encoding": HDF_ENCODING,
    "file_kwds": "{}",
    "global_hash_value": -1,
    "info": "{}",
    "item_type": "dict",
    "length": 0,
    "load_as_complex": "{}",
    "pre_transform": "unknown",
    "shape_suffix": SHAPE_SUFFIX,
    "source_dataset": "unknown",
    "src_np_dtypes": "{}",
    "store_complex_as_real": False,
    "store_str_as_vlen": False,
    "user_attrs": "None",
    "torchoutil_version": "unknown",
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import platform
import sys
from pathlib import Path
from typing import Any, Dict

import torch

import torchoutil
from torchoutil.core.packaging import _EXTRA_AVAILABLE
from torchoutil.pyoutil.collections import dump_dict
from torchoutil.utils.data.dataloader import get_auto_num_cpus, get_auto_num_gpus


def get_package_repository_path() -> str:
    """Return the absolute path where the source code of this package is installed."""
    return str(Path(__file__).parent.parent.parent)


def get_install_info() -> Dict[str, Any]:
    install_info = {
        "torchoutil": torchoutil.__version__,
        "python": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        "os": platform.system(),
        "architecture": platform.architecture()[0],
        "torch": str(torch.__version__),
        "package_path": get_package_repository_path(),
        "num_cpus": get_auto_num_cpus(),
        "num_gpus": get_auto_num_gpus(),
    }
    install_info.update({f"{k}_available": str(v) for k, v in _EXTRA_AVAILABLE.items()})
    return install_info


def print_install_info() -> None:
    """Show main packages versions."""
    install_info = get_install_info()
    dumped = dump_dict(install_info, join="\n", fmt="{key}: {value}")
    print(dumped)


if __name__ == "__main__":
    print_install_info()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, Union

from torch import Tensor, nn

from torchoutil.nn.functional.activation import log_softmax_multidim, softmax_multidim
from torchoutil.pyoutil.collections import dump_dict


class SoftmaxMultidim(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.activation.softmax_multidim`.
    """

    def __init__(
        self,
        dims: Union[Iterable[int], None] = (-1,),
    ) -> None:
        super().__init__()
        self.dims = dims

    def forward(
        self,
        input: Tensor,
    ) -> Tensor:
        return softmax_multidim(input, dims=self.dims)

    def extra_repr(self) -> str:
        return dump_dict(dict(dims=self.dims))


class LogSoftmaxMultidim(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.activation.softmax_multidim`.
    """

    def __init__(
        self,
        dims: Union[Iterable[int], None] = (-1,),
    ) -> None:
        super().__init__()
        self.dims = dims

    def forward(
        self,
        input: Tensor,
    ) -> Tensor:
        return log_softmax_multidim(input, dims=self.dims)

    def extra_repr(self) -> str:
        return dump_dict(dict(dims=self.dims))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Any, Optional, Union

from torchoutil.pyoutil.json import load_json  # noqa: F401
from torchoutil.pyoutil.json import to_json as to_json_base
from torchoutil.utils.saving.common import to_builtin


def to_json(
    data: Any,
    fpath: Union[str, Path, None] = None,
    *,
    overwrite: bool = True,
    make_parents: bool = True,
    to_builtins: bool = False,
    # JSON dump kwargs
    indent: Optional[int] = 4,
    ensure_ascii: bool = False,
    **json_dump_kwds,
) -> str:
    """Dump content to csv format."""
    if to_builtins:
        data = to_builtin(data)

    return to_json_base(
        data,
        fpath,
        overwrite=overwrite,
        make_parents=make_parents,
        indent=indent,
        ensure_ascii=ensure_ascii,
        **json_dump_kwds,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
import unittest
from unittest import TestCase

import torch

from torchoutil.types import tensor_subclasses
from torchoutil.types.tensor_subclasses import (
    BoolTensor,
    BoolTensor1D,
    ByteTensor,
    CDoubleTensor,
    CFloatTensor,
    CHalfTensor,
    ComplexFloatingTensor,
    ComplexFloatingTensor1D,
    DoubleTensor,
    FloatingTensor,
    FloatingTensor0D,
    FloatTensor,
    FloatTensor0D,
    FloatTensor1D,
    FloatTensor2D,
    HalfTensor,
    IntTensor,
    IntTensor0D,
    LongTensor,
    LongTensor0D,
    LongTensor1D,
    ShortTensor,
    ShortTensor1D,
    SignedIntegerTensor,
    SignedIntegerTensor3D,
    Tensor0D,
    Tensor1D,
    UnsignedIntegerTensor,
    _TensorNDBase,
)


class TestTensorTyping(TestCase):
    def test_instance_checks(self) -> None:
        x = torch.rand((), dtype=torch.float)

        assert isinstance(x, torch.Tensor)
        assert isinstance(x, Tensor0D)
        assert isinstance(x, FloatTensor)
        assert isinstance(x, FloatTensor0D)

        assert not isinstance(x, Tensor1D)
        assert not isinstance(x, IntTensor)
        assert not isinstance(x, IntTensor0D)
        assert not isinstance(x, str)
        assert not isinstance(x, float)

    def test_subclassing_checks(self) -> None:
        assert issubclass(Tensor0D, torch.Tensor)
        assert issubclass(FloatTensor, torch.Tensor)
        assert issubclass(FloatTensor0D, torch.Tensor)

        assert issubclass(FloatTensor0D, Tensor0D)
        assert issubclass(FloatTensor0D, FloatTensor)

        assert not issubclass(FloatTensor0D, IntTensor0D)
        assert not issubclass(FloatTensor0D, FloatTensor1D)
        assert not issubclass(FloatTensor0D, str)
        assert not issubclass(FloatTensor0D, float)

        assert not issubclass(torch.Tensor, FloatTensor)
        assert not issubclass(torch.Tensor, Tensor0D)
        assert not issubclass(torch.Tensor, FloatTensor0D)
        assert not issubclass(torch.Tensor, IntTensor)

        assert issubclass(ByteTensor, UnsignedIntegerTensor)
        assert issubclass(BoolTensor, UnsignedIntegerTensor)

        assert not issubclass(ByteTensor, SignedIntegerTensor)
        assert not issubclass(ByteTensor, FloatingTensor)
        assert not issubclass(ByteTensor, ComplexFloatingTensor)

        assert not issubclass(BoolTensor, SignedIntegerTensor)
        assert not issubclass(BoolTensor, FloatingTensor)
        assert not issubclass(BoolTensor, ComplexFloatingTensor)

    def test_default_dtype(self) -> None:
        x = [random.randint(0, 100) for _ in range(10)]

        assert Tensor1D(x).dtype == torch.long  # torch.long is the inferred dtype for x
        assert FloatTensor(x).dtype == torch.float
        assert DoubleTensor(x).dtype == torch.double
        assert LongTensor(x).dtype == torch.long

    def test_torch_floattensor_compat(self) -> None:
        args_lst = [
            (),
            (1, 2),
            (0,),
            (0, 0),
            (1, 0, 2),
            ([1, 2, 3],),
            ([],),
        ]
        for args in args_lst:
            x = DoubleTensor(*args)
            x_pt = torch.DoubleTensor(*args)

            assert isinstance(x, DoubleTensor)
            assert isinstance(x, torch.DoubleTensor)
            assert isinstance(x_pt, DoubleTensor)
            assert isinstance(x_pt, torch.DoubleTensor)

            assert (
                x.shape == x_pt.shape
                and x.dtype == x_pt.dtype
                and x.device == x_pt.device
            ), f"{args=}"

    def test_instantiation_no_args(self) -> None:
        x = Tensor0D()
        assert x.dtype == torch.get_default_dtype() and x.ndim == 0

        x = FloatTensor()
        assert x.dtype == torch.float

        x = FloatTensor0D()
        assert x.dtype == torch.float and x.ndim == 0

        x = FloatTensor1D()
        assert x.dtype == torch.float and x.ndim == 1

        x = LongTensor()
        assert x.dtype == torch.long

        x = LongTensor0D()
        assert x.dtype == torch.long and x.ndim == 0

        x = LongTensor1D()
        assert x.dtype == torch.long and x.ndim == 1

        x = ComplexFloatingTensor1D()
        assert (
            x.ndim == 1
            and not x.is_floating_point()
            and x.is_complex()
            and x.is_signed()
        )

        x = FloatingTensor0D()
        assert (
            x.ndim == 0
            and x.is_floating_point()
            and not x.is_complex()
            and x.is_signed()
        )

        x = SignedIntegerTensor3D()
        assert (
            x.ndim == 3
            and not x.is_floating_point()
            and not x.is_complex()
            and x.is_signed()
        )

    def test_instantiation_with_args(self) -> None:
        args = (1, 2, 3)
        kwargs = dict()
        x = LongTensor(*args, **kwargs)
        assert x.ndim == len(args) and x.dtype == torch.long

        with self.assertRaises(ValueError):
            FloatTensor1D([1, 2], dtype=torch.float16)

        with self.assertRaises(ValueError):
            FloatTensor1D(1, 2)

        with self.assertRaises(ValueError):
            FloatTensor2D([2], 1)  # type: ignore

        with self.assertRaises(ValueError):
            Tensor0D([1, 2], dtype=torch.float16)

        with self.assertRaises(ValueError):
            ShortTensor1D([1, 2], dtype=torch.uint8)

        with self.assertRaises(ValueError):
            ComplexFloatingTensor(dtype=torch.float32)

        with self.assertRaises(ValueError):
            FloatingTensor(dtype=torch.int32)

        with self.assertRaises(ValueError):
            SignedIntegerTensor(dtype=torch.complex64)

        with self.assertRaises(ValueError):
            SignedIntegerTensor(dtype=torch.bool)

        _ = ComplexFloatingTensor(dtype=torch.complex128)
        _ = FloatingTensor(dtype=torch.half)
        _ = SignedIntegerTensor(dtype=torch.long)

        x = Tensor0D(10.0)
        assert x.ndim == 0 and x.dtype == torch.float

    def test_instantiate_all(self) -> None:
        module = tensor_subclasses
        for name in dir(module):
            elem = getattr(module, name)
            if (
                not isinstance(elem, type)
                or not issubclass(elem, _TensorNDBase)
                or elem is _TensorNDBase
            ):
                continue

            tensor_cls = elem
            tensor = tensor_cls()

            assert isinstance(tensor, torch.Tensor)

    def test_complex_class(self) -> None:
        cls = ComplexFloatingTensor
        assert not cls().is_floating_point()
        assert cls().is_complex()
        assert cls().is_signed()
        assert isinstance(cls(), cls)

        assert issubclass(CFloatTensor, cls)
        assert issubclass(CHalfTensor, cls)
        assert issubclass(CDoubleTensor, cls)

        assert isinstance(torch.rand(10, dtype=torch.complex64), cls)
        assert isinstance(CFloatTensor(), cls)

        assert not isinstance(torch.rand(10), cls)
        assert not isinstance(torch.randint(0, 10, (5,)), cls)
        assert not isinstance(FloatTensor(), cls)
        assert not isinstance(DoubleTensor(), cls)

        assert not isinstance(LongTensor(), cls)

        assert not isinstance(BoolTensor1D(), cls)

    def test_floating_class(self) -> None:
        cls = FloatingTensor
        assert cls().is_floating_point()
        assert not cls().is_complex()
        assert cls().is_signed()
        assert isinstance(cls(), cls)

        assert issubclass(FloatTensor, cls)
        assert issubclass(HalfTensor, cls)
        assert issubclass(DoubleTensor, cls)

        assert not isinstance(torch.rand(10, dtype=torch.complex64), cls)
        assert not isinstance(CFloatTensor(), cls)

        assert isinstance(torch.rand(10), cls)
        assert isinstance(FloatTensor(), cls)
        assert isinstance(DoubleTensor(), cls)

        assert not isinstance(torch.randint(0, 10, (5,)), cls)
        assert not isinstance(LongTensor(), cls)

        assert not isinstance(BoolTensor1D(), cls)

    def test_integral_class(self) -> None:
        cls = SignedIntegerTensor
        assert not cls().is_floating_point()
        assert not cls().is_complex()
        assert cls().is_signed()
        assert isinstance(cls(), cls)

        assert issubclass(IntTensor, cls)
        assert issubclass(ShortTensor, cls)
        assert issubclass(LongTensor, cls)

        assert not isinstance(torch.rand(10, dtype=torch.complex64), cls)
        assert not isinstance(CFloatTensor(), cls)

        assert not isinstance(torch.rand(10), cls)
        assert not isinstance(FloatTensor(), cls)
        assert not isinstance(DoubleTensor(), cls)

        assert isinstance(torch.randint(0, 10, (5,)), cls)
        assert isinstance(LongTensor(), cls)

        assert not isinstance(BoolTensor1D(), cls)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.core.dtype_enum import DTypeEnum


class TestDTypeEnum(TestCase):
    def test_dtype_enum_inst(self) -> None:
        dt1 = DTypeEnum.float16
        assert dt1.dtype == torch.float16 == torch.half
        assert dt1 == "float16"
        assert dt1 == "half"
        assert dt1 == "torch.float16"
        assert dt1 == "torch.half"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from torchoutil.core.packaging import _EXTRA_AVAILABLE

_MISSING_DEPS = {
    name: _EXTRA_AVAILABLE[name]
    for name in ("h5py", "numpy", "tqdm")
    if not _EXTRA_AVAILABLE[name]
}
if len(_MISSING_DEPS) > 0:
    if len(_MISSING_DEPS) == 1:
        deps_msg = f"dependancy '{next(iter(_MISSING_DEPS.keys()))}' is"
    else:
        deps_msg = (
            "dependancies " + ", ".join(f"'{k}'" for k in _MISSING_DEPS.keys()) + " are"
        )

    msg = f"Cannot import hdf objects because optional {deps_msg} not installed. Please install it using 'pip install torchoutil[extras]'"
    raise ImportError(msg)

del _MISSING_DEPS

from .dataset import HDFDataset
from .pack import pack_to_hdf



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import warnings
from unittest import TestCase

import torch

from torchoutil.core.packaging import _NUMPY_AVAILABLE
from torchoutil.extras.numpy import np
from torchoutil.nn.functional.others import (
    all_eq,
    can_be_converted_to_tensor,
    can_be_stacked,
    ndim,
    shape,
)


class TestCanBeConvertedToTensor(TestCase):
    def test_examples(self) -> None:
        examples = [
            [],
            (),
            [[1, 0, 0], [2, 3, 4]],
            [[1, 0, 0], [2, 3]],
            [[[True], [False]], [[False], [True]]],
            [[[]], [[]]],
            [torch.rand(10), torch.rand(10)],
            [torch.rand(10, 5), torch.rand(10, 3)],
            [torch.rand(10, 5), torch.rand(10, 5)],
            [torch.rand(10, 5), torch.rand(10, 5, 3)],
            [[torch.rand(10)]],
            torch.rand(10, 5),
            [(), []],
            [2, []],
            "",
            [[[]], []],
        ]

        if _NUMPY_AVAILABLE:
            warnings.filterwarnings("ignore", category=UserWarning)
            examples += [
                np.float64(42),
                [[np.float64(42)], np.array([2])],
                np.random.rand(10, 5),
                [[np.complex128(42)], np.array([2])],
                [np.float16(42), np.float32(99)],
            ]

        for example in examples:
            try:
                torch.as_tensor(example)
                convertible = True
            except TypeError:
                convertible = False
            except ValueError:
                convertible = False

            try:
                torch.stack(example)
                stackable = True
            except TypeError:
                stackable = False
            except RuntimeError:
                stackable = False

            # note: can_be_converted_to_tensor(example) => convertible, but not necessary equal
            assert (
                not can_be_converted_to_tensor(example) or convertible
            ), f"can_be_converted_to_tensor: {example=}"

            assert can_be_stacked(example) == stackable, f"can_be_stacked: {example=}"


class TestNDimShape(TestCase):
    def test_examples(self) -> None:
        examples = [
            [],
            (),
            [[1, 0, 0], [2, 3, 4]],
            [[1, 0, 0], [2, 3]],
            [[[True], [False]], [[False], [True]]],
            [[[]], [[]]],
            [torch.rand(10), torch.rand(10)],
            [torch.rand(10, 5), torch.rand(10, 3)],
            [torch.rand(10, 5), torch.rand(10, 5)],
            [[torch.rand(10)]],
            torch.rand(10, 5),
            [(), []],
            [2, []],
            "",
            [[[]], []],
            [1, [[], 2]],
        ]
        expected_ndims = [
            1,
            1,
            2,
            2,
            3,
            3,
            2,
            3,
            3,
            3,
            2,
            2,
            ValueError,
            0,
            ValueError,
            ValueError,
        ]
        expected_shapes = [
            (0,),
            (0,),
            (2, 3),
            ValueError,
            (2, 2, 1),
            (2, 1, 0),
            (2, 10),
            ValueError,
            (2, 10, 5),
            (1, 1, 10),
            (10, 5),
            (2, 0),
            ValueError,
            (),
            ValueError,
            ValueError,
        ]

        if _NUMPY_AVAILABLE:
            examples += [
                np.float64(42),
                [[np.float64(42)], np.array([2])],
                np.random.rand(10, 5),
                [[np.complex128(42)], np.array([2])],
                [np.float16(42), np.float32(99)],
            ]
            expected_ndims += [0, 2, 2, 2, 1]
            expected_shapes += [(), (2, 1), (10, 5), (2, 1), (2,)]

        assert len(examples) == len(expected_ndims) == len(expected_shapes)

        for i, (example, expected_ndim, expected_shape) in enumerate(
            zip(examples, expected_ndims, expected_shapes)
        ):
            if isinstance(expected_ndim, int):
                result = ndim(example)
                assert result == expected_ndim, f"{example=}"
            else:
                with self.assertRaises(expected_ndim):
                    invalid_ndim = ndim(example)
                    print(f"{invalid_ndim=} with {example=} at {i=}")

            if isinstance(expected_shape, tuple):
                result = shape(example)
                assert result == expected_shape, f"{example=}"
            else:
                with self.assertRaises(expected_shape):
                    invalid_shape = shape(example)
                    print(f"{invalid_shape=} with {example=} at {i=}")


class TestAllEq(TestCase):
    def test_all_eq_example(self) -> None:
        x = torch.full((100, 2, 4), torch.rand(()).item())
        assert all_eq(x)

    def test_all_eq_dim(self) -> None:
        x = torch.as_tensor(
            [
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ]
        )

        assert not all_eq(x)
        assert all_eq(x, dim=0).all()
        assert not all_eq(x, dim=1).any()

        assert not all_eq(x.T)
        assert not all_eq(x.T, dim=0).any()
        assert all_eq(x.T, dim=1).all()

        if _NUMPY_AVAILABLE:
            x = x.numpy()
            assert not all_eq(x)
            assert all_eq(x, dim=0).all()
            assert not all_eq(x, dim=1).any()

            assert not all_eq(x.T)
            assert not all_eq(x.T, dim=0).any()
            assert all_eq(x.T, dim=1).all()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Literal, Optional, Union

import torch
from torch import Generator
from torch.types import Device

from .dtype_enum import DTypeEnum, enum_dtype_to_torch_dtype, str_to_torch_dtype

DeviceLike = Union[Device, Literal["cuda_if_available"]]
DTypeLike = Union[torch.dtype, None, Literal["default"], str, DTypeEnum]
GeneratorLike = Union[int, Generator, None, Literal["default"]]

CUDA_IF_AVAILABLE = "cuda_if_available"


def get_device(device: DeviceLike = CUDA_IF_AVAILABLE) -> Optional[torch.device]:
    if isinstance(device, (torch.device, type(None))):
        return device
    elif device == CUDA_IF_AVAILABLE:
        return torch.device("cuda" if torch.cuda.is_available() else "cpu")
    elif isinstance(device, (str, int)):
        device = torch.device(device)
    else:
        msg = f"Invalid argument type {type(device)}. (expected torch.device, None, str or int)"
        raise TypeError(msg)


def get_dtype(dtype: DTypeLike = None) -> Optional[torch.dtype]:
    if isinstance(dtype, (torch.dtype, type(None))):
        return dtype
    elif dtype == "default":
        return torch.get_default_dtype()
    elif isinstance(dtype, DTypeEnum):
        return enum_dtype_to_torch_dtype(dtype)
    elif isinstance(dtype, str):
        return str_to_torch_dtype(dtype)
    else:
        msg = f"Invalid argument type {type(dtype)}. (expected torch.dtype, None, str or torchoutil.DTypeEnum)"
        raise TypeError(msg)


def get_generator(generator: GeneratorLike = None) -> Optional[Generator]:
    if isinstance(generator, (Generator, type(None))):
        return generator
    elif isinstance(generator, int):
        return Generator().manual_seed(generator)
    elif generator == "default":
        return torch.default_generator
    else:
        msg = f"Invalid argument type {type(generator)}. (expected torch.Generator, None, int or 'default')"
        raise TypeError(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Generic, Iterable, Tuple, TypeVar, Union

import torch
from torch import Tensor

import torchoutil as to
from torchoutil import pyoutil as po
from torchoutil.extras.numpy.definitions import ACCEPTED_NUMPY_DTYPES, np
from torchoutil.pyoutil import BuiltinScalar, get_current_fn_name

T_Invalid = TypeVar("T_Invalid", covariant=True)
T_EmptyNp = TypeVar("T_EmptyNp", covariant=True)
T_EmptyTorch = TypeVar("T_EmptyTorch", covariant=True)


class InvalidTorchDType(metaclass=po.Singleton):
    """Default return type for torch_dtype when an invalid data is passed as argument of scan_torch_dtype function. (like str for example)"""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


@dataclass(frozen=True)
class ShapeDTypeInfo(Generic[T_Invalid, T_EmptyTorch, T_EmptyNp]):
    shape: Tuple[int, ...]
    torch_dtype: Union[torch.dtype, T_Invalid, T_EmptyTorch]
    numpy_dtype: Union[np.dtype, T_EmptyNp]
    valid_shape: bool

    @property
    def fill_value(self) -> BuiltinScalar:
        return numpy_dtype_to_fill_value(self.numpy_dtype)

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @property
    def kind(self) -> str:
        if isinstance(self.numpy_dtype, np.dtype):
            return self.numpy_dtype.kind
        else:
            return "V"


def scan_shape_dtypes(
    x: Any,
    *,
    accept_heterogeneous_shape: bool = False,
    empty_torch: T_EmptyTorch = None,
    empty_np: T_EmptyNp = np.dtype("V"),
) -> ShapeDTypeInfo[InvalidTorchDType, T_EmptyTorch, T_EmptyNp]:
    """Returns the shape and the hdf_dtype for an input."""
    valid_shape, shape = to.shape(x, return_valid=True)
    if not accept_heterogeneous_shape and not valid_shape:
        msg = f"Invalid argument {x} for {get_current_fn_name()}. (cannot compute shape for heterogeneous data)"
        raise ValueError(msg)

    torch_dtype = scan_torch_dtype(x, empty=empty_torch)
    numpy_dtype = scan_numpy_dtype(x, empty=empty_np)

    info = ShapeDTypeInfo[InvalidTorchDType, T_EmptyTorch, T_EmptyNp](
        shape,
        torch_dtype,
        numpy_dtype,
        valid_shape,
    )
    return info


def scan_torch_dtype(
    x: Any,
    *,
    invalid: T_Invalid = InvalidTorchDType(),
    empty: T_EmptyTorch = None,
) -> Union[torch.dtype, T_Invalid, T_EmptyTorch]:
    """Returns torch dtype of an arbitrary object. Works recursively on tuples and lists. An instance of InvalidTorchDType can be returned if a str is passed."""
    if isinstance(x, (int, float, bool, complex)):
        torch_dtype = torch.as_tensor(x).dtype
        return torch_dtype

    if isinstance(x, Tensor):
        torch_dtype = x.dtype
        return torch_dtype

    if isinstance(x, (np.ndarray, np.generic)):
        torch_dtype = numpy_dtype_to_torch_dtype(x.dtype, invalid=invalid)
        return torch_dtype

    if isinstance(x, (str, bytes, bytearray)):
        return invalid

    if isinstance(x, (list, tuple)):
        if len(x) == 0:
            return empty

        torch_dtypes = [scan_torch_dtype(xi, invalid=invalid, empty=empty) for xi in x]
        torch_dtype = merge_torch_dtypes(torch_dtypes, invalid=invalid, empty=empty)
        return torch_dtype

    msg = f"Unsupported type {x.__class__.__name__} in function {po.get_current_fn_name()}."
    raise TypeError(msg)


def scan_numpy_dtype(
    x: Any,
    *,
    empty: T_EmptyNp = np.dtype("V"),
) -> Union[np.dtype, T_EmptyNp]:
    if isinstance(x, (int, float, bool, complex)):
        numpy_dtype = np.array(x).dtype
        return numpy_dtype

    if isinstance(x, Tensor):
        numpy_dtype = torch_dtype_to_numpy_dtype(x.dtype)
        return numpy_dtype

    if isinstance(x, (np.ndarray, np.generic)):
        numpy_dtype = x.dtype
        return numpy_dtype

    if isinstance(x, (str, bytes, bytearray)):
        numpy_dtype = np.array(x).dtype
        return numpy_dtype

    if isinstance(x, (list, tuple)):
        if len(x) == 0:
            return empty

        numpy_dtypes = [scan_numpy_dtype(xi, empty=empty) for xi in x]
        numpy_dtype = merge_numpy_dtypes(numpy_dtypes, empty=empty)
        return numpy_dtype

    msg = f"Unsupported type {x.__class__.__name__} in function {po.get_current_fn_name()}."
    raise TypeError(msg)


@lru_cache(maxsize=None)
def torch_dtype_to_numpy_dtype(dtype: torch.dtype) -> np.dtype:
    x = torch.empty((0,), dtype=dtype)
    x = to.tensor_to_numpy(x)
    return x.dtype


@lru_cache(maxsize=None)
def numpy_dtype_to_torch_dtype(
    dtype: np.dtype,
    *,
    invalid: T_Invalid = InvalidTorchDType(),
) -> Union[torch.dtype, T_Invalid]:
    if dtype in ACCEPTED_NUMPY_DTYPES:
        x = np.empty((0,), dtype=dtype)
        x = to.numpy_to_tensor(x)
        return x.dtype
    else:
        return invalid


def numpy_dtype_to_fill_value(dtype: Any) -> BuiltinScalar:
    if not isinstance(dtype, np.dtype):
        return None

    kind = dtype.kind
    if kind in ("b",):
        return False
    elif kind in ("u", "i"):
        return 0
    elif kind in ("f",):
        return 0.0
    elif kind in ("c",):
        return 0j
    elif kind in ("U", "S"):
        return ""
    else:
        KINDS = ("b", "u", "i", "f", "c", "U", "S")
        msg = f"Invalid argument {dtype=}. (expected dtype.kind in {KINDS})"
        raise ValueError(msg)


def merge_numpy_dtypes(
    dtypes: Iterable[Union[np.dtype, T_EmptyNp]],
    *,
    empty: T_EmptyNp = np.dtype("V"),
) -> Union[np.dtype, T_EmptyNp]:
    dtypes = list(dict.fromkeys(dtypes))
    dtypes = [dtype for dtype in dtypes if dtype != empty]
    if len(dtypes) == 0:
        return empty

    dummy_arrays = [np.empty((0,), dtype=dtype) for dtype in dtypes]  # type: ignore
    dtype = np.stack(dummy_arrays).dtype
    return dtype


def merge_torch_dtypes(
    dtypes: Iterable[Union[torch.dtype, T_Invalid, T_EmptyNp]],
    *,
    invalid: T_Invalid = InvalidTorchDType(),
    empty: T_EmptyNp = None,
) -> Union[torch.dtype, T_Invalid, T_EmptyNp]:
    dtypes = list(dict.fromkeys(dtypes))
    dtypes = [dtype for dtype in dtypes if dtype != empty]
    if len(dtypes) == 0:
        return empty
    if any(dtype == invalid for dtype in dtypes):
        return invalid

    dummy_tensors = [torch.empty((0,), dtype=dtype) for dtype in dtypes]  # type: ignore
    dtype = torch.stack(dummy_tensors).dtype
    return dtype


def get_default_numpy_dtype() -> np.dtype:
    return np.empty((0,)).dtype



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, List, Optional, Union

import torch
from torch import LongTensor, Tensor

from torchoutil.core.get import DeviceLike, DTypeLike, get_device, get_dtype


def masked_mean(
    x: Tensor,
    non_pad_mask: Tensor,
    *,
    dim: Union[None, int, Iterable[int]] = None,
) -> Tensor:
    """Average a tensor along the specified dim(s).

    Args:
        tensor: (N, ...)
        non_pad_mask: Non-padding mask, should be broadcastable with argument tensor and reduced with argument dim.
            It should be a boolean tensor or a float tensor containing only 1 and 0 values.
        dim: Optional dim(s) to reduce. If None, result will be reduced to a scalar. defaults to None.
    """
    if dim is None:
        dim = ()
    elif isinstance(dim, int):
        dim = (dim,)
    else:
        dim = tuple(dim)

    masked = x * non_pad_mask
    reduced = masked.sum(dim=dim) / non_pad_mask.sum(dim=dim).clamp(min=1.0)
    return reduced


def masked_sum(
    x: Tensor,
    non_pad_mask: Tensor,
    *,
    dim: Union[None, int, Iterable[int]] = None,
) -> Tensor:
    """Sum a tensor along the specified dim(s).

    Args:
        x: (N, ...)
        non_pad_mask: Non-padding mask, should be broadcastable with argument tensor and reduced with argument dim.
            It should be a boolean tensor or a float tensor containing only 1 and 0 values.
        dim: Optional dim(s) to reduce. If None, result will be reduced to a scalar. defaults to None.
    """
    if dim is None:
        dim = ()
    elif isinstance(dim, int):
        dim = (dim,)
    else:
        dim = tuple(dim)

    masked = x * non_pad_mask
    reduced = masked.sum(dim=dim)
    return reduced


def masked_equal(
    x1: Tensor,
    x2: Tensor,
    mask: Tensor,
) -> bool:
    """Check if two tensors are equal at the specific positions.

    Args:
        x1: First tensor of shape S.
        x2: Second tensor of shape S.
        mask: Boolean tensor of shape S. Position marked as False are ignored by the equality.
    """
    if x1.shape != x2.shape:
        return False
    mask = mask.bool().logical_not().logical_or(x1.eq(x2))
    equal = mask.all().item()
    return equal  # type: ignore


def generate_square_subsequent_mask(
    size: int,
    diagonal: int = 0,
    *,
    device: DeviceLike = None,
    dtype: DTypeLike = None,
) -> Tensor:
    device = get_device(device)
    dtype = get_dtype(dtype)

    mask = torch.ones((size, size), device=device, dtype=torch.bool)
    mask = torch.tril(mask, diagonal=diagonal)
    mask = torch.where(mask, 0.0, -torch.inf)
    mask = mask.to(dtype=dtype)

    return mask


def lengths_to_non_pad_mask(
    lengths: Tensor,
    max_len: Optional[int] = None,
    include_end: bool = False,
    *,
    dtype: DTypeLike = None,
) -> Tensor:
    """Convert lengths to binary mask of non-padded values.

    The output will be a tensor of shape (B, max_len).

    Args:
        lengths: (bsize,)
        max_len: Optional int for indicate the maximal length.
            If None, it will be set to lengths.max().
            defaults to None.
        include_end: If True, the value at index of len will be True in returned mask.
            defaults to False.

    Example 1::
    -----------
        >>> input = torch.as_tensor([4, 2, 0, 3, 0])
        >>> lengths_to_non_pad_mask(input, max_len=6, include_end=False)
        tensor([[True, True, True, True, False, False],
                [True, True, False, False, False, False],
                [False, False, False, False, False, False],
                [True, True, True, False, False, False],
                [False, False, False, False, False, False]])
    """
    dim = -1
    if max_len is None:
        max_len = int(lengths.max(dim=dim)[0].item())
    indices = torch.arange(0, max_len, device=lengths.device)
    lengths = lengths.unsqueeze(dim=-1)
    if include_end:
        non_pad_mask = indices <= lengths
    else:
        non_pad_mask = indices < lengths
    dtype = get_dtype(dtype)
    non_pad_mask = non_pad_mask.to(dtype=dtype)
    return non_pad_mask


def lengths_to_pad_mask(
    lengths: Tensor,
    max_len: Optional[int] = None,
    include_end: bool = True,
    *,
    dtype: DTypeLike = None,
) -> Tensor:
    """Convert lengths to binary mask of padded values.
    The output will be a tensor of shape (B, max_len).

    Args:
        lengths: (B,)
        max_len: Optional int for indicate the maximal length.
            If None, it will be set to lengths.max().
            defaults to None.
        include_end: If True, the last value of each size will be set to False.
            defaults to True.

    Example 1::
    -----------
        >>> input = torch.as_tensor([4, 2, 0, 3, 0])
        >>> lengths_to_non_pad_mask(input, max_len=None, include_end=True)
        tensor([[False, False, False, False],
                [False, False, True, True],
                [True, True, True, True],
                [False, False, False, True],
                [True, True, True, True]])
    """
    non_pad_mask = lengths_to_non_pad_mask(
        lengths,
        max_len,
        not include_end,
        dtype=torch.bool,
    )
    pad_mask = non_pad_mask.logical_not()
    dtype = get_dtype(dtype)
    pad_mask = pad_mask.to(dtype=dtype)
    return pad_mask


def non_pad_mask_to_lengths(mask: Tensor, *, dim: int = -1) -> LongTensor:
    return mask.sum(dim=dim)  # type: ignore


def pad_mask_to_lengths(mask: Tensor, *, dim: int = -1) -> LongTensor:
    return mask.shape[dim] - non_pad_mask_to_lengths(mask, dim=dim)  # type: ignore


def tensor_to_lengths(
    tensor: Tensor,
    *,
    pad_value: Optional[float] = None,
    end_value: Optional[float] = None,
    dim: int = -1,
) -> LongTensor:
    """Get the lengths of the non-padded elements of a tensor.

    You must provide a value for one of `pad_value` or `end_value`.
    If both values are provided, the `end_value` is ignored.
    The output will be of shape (N,).
    The `end_value` is not included in the length of the sentence.

    Args:
        tensor: Input of shape (N, *).
        pad_value: The pad value used in `tensor`. defaults to None.
        end_value: The end value used in `tensor`. defaults to None.
        dim: The dimension of the length. defaults to -1.

    Example 1::
    -----------
    ```
    >>> x = torch.as_tensor([1, 10, 20, 2, 0, 0])
    >>> tensor_to_lengths(x, end_value=2)
    ... tensor(3)
    ```

    Example 2::
    -----------
    ```
    >>> x = torch.as_tensor([1, 10, 20, 2, 0, 0])
    >>> tensor_to_lengths(x, pad_value=0)
    ... tensor(4)
    ```

    """
    if (pad_value is None) == (end_value is None):
        msg = "Invalid arguments. Please provide only one of the arguments: end_value, pad_value."
        raise ValueError(msg)

    if pad_value is not None:
        non_pad_mask = tensor != pad_value
        lengths = non_pad_mask.sum(dim=dim)

    elif end_value is not None:
        contains_eos = (tensor == end_value).any(dim=dim)
        indices_eos = (tensor == end_value).int().argmax(dim=dim)
        lengths = torch.where(contains_eos, indices_eos, tensor.shape[dim])

    else:
        msg = "Invalid arguments. Please provide only one of the arguments : end_value, pad_value."
        raise ValueError(msg)

    return lengths  # type: ignore


def tensor_to_non_pad_mask(
    tensor: Tensor,
    *,
    pad_value: Optional[float] = None,
    end_value: Optional[float] = None,
    include_end: bool = False,
    dtype: DTypeLike = None,
) -> Tensor:
    """Convert tensor to non-pad binary mask.
    You must provide a value for one of pad_value or end_value. If both values are provided, the end_value is ignored.
    The output will be a binary mask representing the non-padded values. Shape is the same than the input tensor.

    Args:
        tensor: A tensor of values. If end_value is given instead of pad_value, the number of dims must be <= 2.
        pad_value: The pad value used in tensor. defaults to None.
        end_value: The end value used in tensor. defaults to None.
        include_end: If True, the end value will be included in non_pad_mask.
            This parameter is ignored if end_value is None.
            defaults to False.

    Example 1::
    -----------
        >>> input = torch.as_tensor([1, 10, 20, 2, 0, 0])
        >>> tensor_to_pad_mask(input, end_value=2)
        tensor([True, True, True, False, False, False])
    """
    dtype = get_dtype(dtype)

    if (pad_value is None) == (end_value is None):
        msg = "Invalid arguments. Please provide only one of the arguments: end_value, pad_value."
        raise ValueError(msg)

    if pad_value is not None:
        non_pad_mask = tensor.ne(pad_value)
        non_pad_mask = non_pad_mask.to(dtype=dtype)

    elif end_value is not None:
        if tensor.ndim > 2:
            msg = f"Cannot compute non_pad_mask for with more than 2 dimensions with {end_value=}. (found {tensor.ndim=})"
            raise ValueError(msg)
        lengths = tensor_to_lengths(tensor, end_value=end_value, dim=-1)
        non_pad_mask = lengths_to_non_pad_mask(
            lengths, tensor.shape[-1], include_end, dtype=dtype
        )

    else:
        msg = "Invalid arguments. Please provide only one of the arguments : end_value, pad_value."
        raise ValueError(msg)

    return non_pad_mask


def tensor_to_pad_mask(
    tensor: Tensor,
    *,
    pad_value: Optional[float] = None,
    end_value: Optional[float] = None,
    include_end: bool = True,
    dtype: DTypeLike = None,
) -> Tensor:
    """Convert tensor to pad binary mask.

    You must provide a value for one of pad_value or end_value. If both values are provided, the end_value is ignored.

    The output will be a binary mask representing the padded values. Shape is the same than the input tensor.

    Args:
        tensor: A tensor of values. If end_value is given instead of pad_value, the number of dims must be <= 2.
        pad_value: The pad value used in tensor. defaults to None.
        end_value: The end value used in tensor. defaults to None.
        include_end: If True, the end value will be included in pad_mask. defaults to True.

    Example 1::
    -----------
        >>> input = torch.as_tensor([1, 10, 20, 2, 0, 0])
        >>> tensor_to_pad_mask(input, end_value=2)
        tensor([False, False, False, True, True, True])
    """
    dtype = get_dtype(dtype)
    non_pad_mask = tensor_to_non_pad_mask(
        tensor,
        pad_value=pad_value,
        end_value=end_value,
        include_end=not include_end,
        dtype=torch.bool,
    )
    pad_mask = non_pad_mask.logical_not()
    pad_mask = pad_mask.to(dtype=dtype)
    return pad_mask


def tensor_to_tensors_list(
    tensor: Tensor,
    *,
    pad_value: Optional[float] = None,
    end_value: Optional[float] = None,
    non_pad_mask: Optional[Tensor] = None,
    lengths: Optional[Tensor] = None,
) -> List[Tensor]:
    """Convert padded tensor to tensor list.

    You must provide a value for one of the 4 arguments: `pad_value`, `end_value`, `non_pad_mask` or `lengths`.
    If multiple values are provided, only one will be used and the priority order is `pad_value`, `end_value`, `non_pad_mask` and `lengths`.
    The output will be a list of N tensors of shape (*).

    Args:
        `tensor`: (N, *)
        `pad_value`: Pad value index. defaults to None.
        `end_value`: End value index. defaults to None.
        `non_pad_mask`: Optional non-padded boolean mask. defaults to None.
        `lengths`: Length of each sequence in padded batch.
    """
    if pad_value is not None:
        lengths = tensor_to_lengths(tensor, pad_value=pad_value)
        return tensor_to_tensors_list(tensor, lengths=lengths)

    elif end_value is not None:
        lengths = tensor_to_lengths(tensor, end_value=end_value)
        return tensor_to_tensors_list(tensor, lengths=lengths)

    elif non_pad_mask is not None:
        lengths = non_pad_mask_to_lengths(non_pad_mask)
        return tensor_to_tensors_list(tensor, lengths=lengths)

    elif lengths is not None:
        slices_lst = [
            [slice(None) for _ in range(tensor.ndim)] + [slice(0, len_)]
            for len_ in lengths
        ]
        tensors = [tensor[slices] for slices in slices_lst]

    else:
        msg = "Invalid arguments. Please provide only one of the arguments : end_value, pad_value, non_pad_mask or lengths."
        raise ValueError(msg)

    return tensors


def tensors_list_to_lengths(tensors: List[Tensor], dim: int = -1) -> LongTensor:
    """Return the size of the tensor at a specific dim.

    The output will be a tensor of size N.

    Args:
        tensors: List of N tensors.
        dim: The dimension of the output sizes. defaults to -1.
    """
    device = None if len(tensors) == 0 else tensors[0].device
    lst = [tensor.shape[dim] for tensor in tensors]
    output = torch.as_tensor(lst, dtype=torch.long, device=device)
    return output  # type: ignore



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from pathlib import Path
from typing import Any, Optional, Union


def to_json(
    data: Any,
    fpath: Union[str, Path, None] = None,
    *,
    overwrite: bool = True,
    make_parents: bool = True,
    # JSON dump kwargs
    indent: Optional[int] = 4,
    ensure_ascii: bool = False,
    **json_dump_kwds,
) -> str:
    """Dump content to JSON format."""
    if fpath is not None:
        fpath = Path(fpath).resolve().expanduser()
        if not overwrite and fpath.exists():
            raise FileExistsError(f"File {fpath} already exists.")
        elif make_parents:
            fpath.parent.mkdir(parents=True, exist_ok=True)

    content = json.dumps(
        data,
        indent=indent,
        ensure_ascii=ensure_ascii,
        **json_dump_kwds,
    )
    if fpath is not None:
        fpath.write_text(content)
    return content


def load_json(fpath: Union[str, Path], **json_load_kwds) -> Any:
    fpath = Path(fpath)
    content = fpath.read_text()
    data = json.loads(content, **json_load_kwds)
    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, List, Optional, Union

DEFAULT_TRUE_VALUES = ("True", "t", "yes", "y", "1")
DEFAULT_FALSE_VALUES = ("False", "f", "no", "n", "0")
DEFAULT_NONE_VALUES = ("None", "null")


def str_to_bool(
    x: str,
    *,
    case_sensitive: bool = False,
    true_values: Union[str, Iterable[str]] = DEFAULT_TRUE_VALUES,
    false_values: Union[str, Iterable[str]] = DEFAULT_FALSE_VALUES,
) -> bool:
    """Convert string values to bool. Intended for argparse boolean arguments.

    - True values: 'True', 'T', 'yes', 'y', '1'.
    - False values: 'False', 'F', 'no', 'n', '0'.
    - Other raises ValueError.
    """
    true_values = _sanitize_values(true_values)
    if _str_in(x, true_values, case_sensitive):
        return True

    false_values = _sanitize_values(false_values)
    if _str_in(x, false_values, case_sensitive):
        return False

    values = tuple(true_values + false_values)
    raise ValueError(f"Invalid argument '{x}'. (expected one of {values})")


def str_to_optional_int(
    x: str,
    *,
    case_sensitive: bool = False,
    none_values: Union[str, Iterable[str]] = DEFAULT_NONE_VALUES,
) -> Optional[int]:
    none_values = _sanitize_values(none_values)
    if _str_in(x, none_values, case_sensitive):
        return None

    return int(x)


def str_to_optional_str(
    x: str,
    *,
    case_sensitive: bool = False,
    none_values: Union[str, Iterable[str]] = DEFAULT_NONE_VALUES,
) -> Optional[str]:
    none_values = _sanitize_values(none_values)
    if _str_in(x, none_values, case_sensitive):
        return None

    return x


def str_to_optional_bool(
    x: str,
    *,
    case_sensitive: bool = False,
    true_values: Union[str, Iterable[str]] = DEFAULT_TRUE_VALUES,
    false_values: Union[str, Iterable[str]] = DEFAULT_FALSE_VALUES,
    none_values: Union[str, Iterable[str]] = DEFAULT_NONE_VALUES,
) -> Optional[bool]:
    true_values = _sanitize_values(true_values)
    if _str_in(x, true_values, case_sensitive):
        return True

    false_values = _sanitize_values(false_values)
    if _str_in(x, false_values, case_sensitive):
        return False

    none_values = _sanitize_values(none_values)
    if _str_in(x, none_values, case_sensitive):
        return None

    values = tuple(true_values + false_values + none_values)
    raise ValueError(f"Invalid argument '{x}'. (expected one of {values})")


def _sanitize_values(values: Union[str, Iterable[str]]) -> List[str]:
    if isinstance(values, str):
        values = [values]
    else:
        values = list(values)
    return values


def _str_in(x: str, values: List[str], case_sensitive: bool) -> bool:
    if case_sensitive:
        return x in values
    else:
        return x.lower() in map(str.lower, values)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from pathlib import Path
from typing import Optional, Union
from urllib.error import HTTPError

from torch.hub import download_url_to_file

from torchoutil.pyoutil.hashlib import hash_file  # noqa: F401
from torchoutil.pyoutil.os import safe_rmdir  # noqa: F401

pylog = logging.getLogger(__name__)


def download_file(
    url: str,
    fpath: Union[str, Path],
    *,
    hash_prefix: Optional[str] = None,
    make_intermediate: bool = False,
    verbose: int = 0,
) -> Path:
    fpath = Path(fpath)
    if make_intermediate:
        dpath = fpath.parent
        dpath.mkdir(parents=True, exist_ok=True)

    try:
        download_url_to_file(
            url,
            fpath,  # type: ignore
            hash_prefix=hash_prefix,
            progress=verbose > 0,
        )

    except HTTPError as err:
        msg = f"Cannot download from {url=}. (with {fpath=}, {hash_prefix=}, {make_intermediate=})"
        pylog.error(msg)
        raise err

    return fpath



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import os
from dataclasses import asdict
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import h5py
import numpy as np
import torch
import tqdm
from h5py import Dataset as HDFRawDataset
from torch.utils.data.dataloader import DataLoader

import pyoutil as po
import torchoutil as to
from torchoutil import nn
from torchoutil.extras.hdf.common import (
    _DUMPED_JSON_KEYS,
    HDF_ENCODING,
    HDF_STRING_DTYPE,
    HDF_VOID_DTYPE,
    SHAPE_SUFFIX,
    HDFItemType,
)
from torchoutil.extras.hdf.dataset import HDFDataset
from torchoutil.extras.numpy import (
    merge_numpy_dtypes,
    numpy_is_complex_dtype,
    scan_shape_dtypes,
)
from torchoutil.pyoutil.collections import all_eq
from torchoutil.pyoutil.datetime import now_iso
from torchoutil.pyoutil.functools import Compose
from torchoutil.pyoutil.typing import is_dataclass_instance, is_dict_str
from torchoutil.types import BuiltinScalar
from torchoutil.utils.data.dataloader import get_auto_num_cpus
from torchoutil.utils.data.dataset import IterableDataset, SizedDatasetLike
from torchoutil.utils.pack.common import EXISTS_MODES, ExistsMode, _tuple_to_dict
from torchoutil.utils.saving.common import to_builtin

T = TypeVar("T", covariant=True)
T_DictOrTuple = TypeVar("T_DictOrTuple", tuple, dict, covariant=True)

HDFDType = Union[np.dtype, Literal["b", "i", "u", "f", "c"], type]


pylog = logging.getLogger(__name__)


@torch.inference_mode()
def pack_to_hdf(
    dataset: SizedDatasetLike[T],
    hdf_fpath: Union[str, Path],
    pre_transform: Optional[Callable[[T], T_DictOrTuple]] = None,
    *,
    exists: ExistsMode = "error",
    verbose: int = 0,
    batch_size: int = 32,
    num_workers: Union[int, Literal["auto"]] = "auto",
    shape_suffix: str = SHAPE_SUFFIX,
    store_str_as_vlen: bool = False,
    file_kwds: Optional[Dict[str, Any]] = None,
    ds_kwds: Optional[Dict[str, Any]] = None,
    user_attrs: Any = None,
    skip_scan: bool = False,
) -> HDFDataset[T_DictOrTuple, T_DictOrTuple]:
    """Pack a dataset to HDF file.

    Args:
        dataset: The sized dataset to pack. Must be sized and all items must be of dict type.
            The key of each dictionaries are strings and values can be int, float, str, Tensor, non-empty List[int], non-empty List[float], non-empty List[str].
            If values are tensors or lists, the number of dimensions must be the same for all items in the dataset.
        hdf_fpath: The path to the HDF file.
        pre_transform: The optional transform to apply to audio returned by the dataset BEFORE storing it in HDF file.
            Can be used for deterministic transforms like Resample, LogMelSpectrogram, etc. defaults to None.
        exists: Determine which action should be performed if the target HDF file already exists.
            "overwrite": Replace the target file then pack dataset.
            "skip": Skip this function and returns the packed dataset.
            "error": Raises a ValueError.
        verbose: Verbose level. defaults to 0.
        batch_size: The batch size of the dataloader. defaults to 32.
        num_workers: The number of workers of the dataloader.
            If "auto", it will be set to `len(os.sched_getaffinity(0))`. defaults to "auto".
        shape_suffix: Shape column suffix in HDF file. defaults to "_shape".
        store_str_as_vlen: If True, store strings as variable length string dtype. defaults to False.
        file_kwds: Options given to h5py file object. defaults to None.
        ds_kwds: Keywords arguments passed to the returned HDFDataset instance if the target file already exists and if exists == "skip".
        user_attrs: Additional metadata to add to the hdf file. It must be convertible to JSON with `json.dumps`. defaults to None.
        skip_scan: If True, the input dataset will be considered as fully homogeneous, which means that all columns values contains the same shape and dtype, which will be inferred from the first batch.
            It is meant to skip the first step which scans each dataset item once and speed up packing to HDF file.
            defaults to False.

    Returns:
        hdf_dataset: The target HDF dataset object.
    """
    if not isinstance(dataset, SizedDatasetLike):
        msg = f"Cannot pack to hdf a non-sized-dataset '{dataset.__class__.__name__}'."
        raise TypeError(msg)
    if len(dataset) == 0:
        msg = f"Cannot pack to hdf an empty dataset. (found {len(dataset)=})"
        raise ValueError(msg)

    hdf_fpath = Path(hdf_fpath).resolve()
    if hdf_fpath.exists() and not hdf_fpath.is_file():
        msg = f"Item {hdf_fpath=} exists but it is not a file."
        raise RuntimeError(msg)

    if ds_kwds is None:
        ds_kwds = {}

    if not hdf_fpath.is_file() or exists == "overwrite":
        pass
    elif exists == "skip":
        return HDFDataset(hdf_fpath, **ds_kwds)
    elif exists == "error":
        msg = f"Cannot overwrite file {hdf_fpath}. Please remove it or use exists='overwrite' or exists='skip' option."
        raise ValueError(msg)
    else:
        msg = f"Invalid argument {exists=}. (expected one of {EXISTS_MODES})"
        raise ValueError(msg)

    if file_kwds is None:
        file_kwds = {}

    if num_workers == "auto":
        num_workers = get_auto_num_cpus()
        if verbose >= 2:
            pylog.debug(f"Found num_workers=='auto', set to {num_workers}.")

    if verbose >= 2:
        pylog.debug(f"Start packing data into HDF file '{hdf_fpath}'...")

    # Step 1: First pass to the dataset to build static HDF dataset shapes (much faster for read the resulting file)
    encoding = HDF_ENCODING
    (
        dict_pre_transform,
        item_type,
        max_shapes,
        hdf_dtypes,
        all_eq_shapes,
        src_np_dtypes,
    ) = _scan_dataset(
        dataset=dataset,
        pre_transform=pre_transform,
        batch_size=batch_size,
        num_workers=num_workers,
        verbose=verbose,
        store_str_as_vlen=store_str_as_vlen,
        encoding=encoding,
        skip_scan=skip_scan,
    )

    total = sum(po.prod(shape) for shape in max_shapes.values())
    max_shapes_ratios = {
        attr_name: po.prod(shape) / total for attr_name, shape in max_shapes.items()
    }

    # For debugging purposes
    data = {
        "item_type": item_type,
        "max_shapes": max_shapes,
        "hdf_dtypes": hdf_dtypes,
        "all_eq_shapes": all_eq_shapes,
        "src_np_dtypes": src_np_dtypes,
    }
    data = to_builtin(data)

    with NamedTemporaryFile(
        "w",
        prefix="HDF_scan_results_",
        suffix=".json",
        delete=False,
    ) as file:
        json.dump(data, file)
        scan_results_fpath = Path(file.name)

    creation_date = now_iso()

    with h5py.File(hdf_fpath, "w", **file_kwds) as hdf_file:
        # Step 2: Build hdf datasets in file
        hdf_dsets: Dict[str, HDFRawDataset] = {}

        # Create sub-datasets for main data
        for attr_name, shape in max_shapes.items():
            hdf_dtype = hdf_dtypes.get(attr_name)

            kwargs: Dict[str, Any] = {}
            fill_value = hdf_dtype_to_fill_value(hdf_dtype)
            if fill_value is not None:
                kwargs["fillvalue"] = fill_value

            hdf_ds_shape = (len(dataset),) + shape
            try:
                hdf_dsets[attr_name] = hdf_file.create_dataset(
                    attr_name, hdf_ds_shape, hdf_dtype, **kwargs
                )
            except ValueError as err:
                msg = f"Cannot create hdf dataset {attr_name=} of shape '{hdf_ds_shape}' with dtype '{hdf_dtype}' and {kwargs=}."
                pylog.error(msg)
                raise err

        if verbose >= 2:
            num_scalars = sum(len(hdf_ds.shape) == 1 for hdf_ds in hdf_dsets.values())
            ratio = num_scalars / total
            msg = f"{num_scalars}/{len(hdf_dsets)} column dsets contains a single dim. ({ratio*100:.3f}%)"
            pylog.debug(msg)

            if num_scalars < len(hdf_dsets):
                msg = "Others multidims column dsets are:"
                pylog.debug(msg)

            for attr_name, hdf_ds in hdf_dsets.items():
                if len(hdf_ds.shape) == 1:
                    continue
                ratio = max_shapes_ratios[attr_name]
                msg = f"HDF column dset multidim '{attr_name}' has been built. (with shape={hdf_ds.shape}, nelement_per_item={po.prod(hdf_ds.shape[1:])} ({ratio*100:.3f}%), dtype={hdf_ds.dtype})"
                pylog.debug(msg)

        added_columns: List[str] = []

        # Create sub-datasets for shape data
        for attr_name, shape in max_shapes.items():
            if len(shape) == 0 or all_eq_shapes[attr_name]:
                continue

            shape_name = f"{attr_name}{shape_suffix}"
            raw_dset_shape = len(dataset), len(shape)

            if shape_name not in hdf_dsets:
                pass
            elif hdf_dsets[shape_name].shape == raw_dset_shape:
                continue
            else:
                msg = f"Column {shape_name} already exists in source dataset with a different shape. (found shape={hdf_dsets[shape_name].shape} but expected shape is {raw_dset_shape})"
                raise RuntimeError(msg)

            hdf_dsets[shape_name] = hdf_file.create_dataset(
                shape_name,
                raw_dset_shape,
                np.int32,
                fillvalue=-1,
            )
            added_columns.append(shape_name)

        # Fill sub-datasets with a second pass through the whole dataset
        i = 0
        global_hash_value = 0

        loader = DataLoader(
            dataset,  # type: ignore
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
            collate_fn=nn.Identity(),
            drop_last=False,
            pin_memory=False,
        )

        for _batch_idx, batch in enumerate(
            tqdm.tqdm(
                loader,
                desc="Pack data into HDF...",
                disable=verbose <= 0,
            )
        ):
            batch = [dict_pre_transform(item) for item in batch]

            for item in batch:
                for attr_name, value in item.items():
                    hdf_dset = hdf_dsets[attr_name]
                    shape = to.shape(value)

                    # Check every shape
                    if len(shape) != hdf_dset.ndim - 1:
                        msg = f"Invalid number of dimension in audio (expected {len(shape)}, found {hdf_dset.ndim - 1=})."
                        raise ValueError(msg)

                    # Check dataset size
                    if any(
                        shape_i > dset_shape_i
                        for shape_i, dset_shape_i in zip(shape, hdf_dset.shape[1:])
                    ):
                        msg = f"Resize hdf_dset {attr_name} of shape {tuple(hdf_dset.shape[1:])} with new {shape=}."
                        pylog.error(msg)
                        msg = "INTERNAL ERROR: Cannot resize dataset when pre-computing shapes."
                        raise RuntimeError(msg)

                    # Note: "hdf_dset[slices]" is a generic version of "hdf_dset[i, :shape_0, :shape_1]"
                    slices = (i,) + tuple(slice(shape_i) for shape_i in shape)

                    try:
                        hdf_dset[slices] = value
                    except (TypeError, ValueError, OSError) as err:
                        msg = f"Cannot set data {value} of shape {shape} into {hdf_dset.shape=} ({attr_name=}, {i=}, {slices=}, {value.dtype=} {hdf_dset.dtype=})"
                        pylog.error(msg)
                        raise err

                    # Store original shape if needed
                    shape_name = f"{attr_name}{shape_suffix}"
                    if shape_name in hdf_dsets.keys():
                        hdf_shapes_dset = hdf_dsets[shape_name]
                        hdf_shapes_dset[i] = shape

                    global_hash_value += to.checksum(value)

                i += 1

        # note: HDF cannot save too large int values with too many bits
        global_hash_value = global_hash_value % (2**31)

        if not hasattr(dataset, "info"):
            info = {}
        else:
            info = dataset.info  # type: ignore
            if is_dataclass_instance(info):
                info = asdict(info)
            elif isinstance(info, Mapping):
                info = dict(info.items())  # type: ignore
            else:
                info = {}

        src_np_dtypes_dumped = {
            name: str(merge_numpy_dtypes(np_dtypes, empty=HDF_VOID_DTYPE))
            for name, np_dtypes in src_np_dtypes.items()
        }
        attributes = {
            "added_columns": added_columns,
            "creation_date": creation_date,
            "encoding": encoding,
            "file_kwds": file_kwds,
            "global_hash_value": global_hash_value,
            "info": to_builtin(info),
            "item_type": item_type,
            "length": len(dataset),
            "load_as_complex": {},  # for backward compatibility only
            "pre_transform": pre_transform.__class__.__name__,
            "shape_suffix": shape_suffix,
            "source_dataset": dataset.__class__.__name__,
            "src_np_dtypes": src_np_dtypes_dumped,
            "store_complex_as_real": False,  # for backward compatibility only
            "store_str_as_vlen": store_str_as_vlen,
            "user_attrs": to_builtin(user_attrs),
            "torchoutil_version": str(to.__version__),
        }
        for name in _DUMPED_JSON_KEYS:
            attributes[name] = json.dumps(attributes[name])

        if verbose >= 2:
            dumped_attributes = json.dumps(attributes, indent="\t")
            pylog.debug(f"Saving attributes in HDF file:\n{dumped_attributes}")

        attrs_errors: List[TypeError] = []
        for attr_name, attr_val in attributes.items():
            try:
                hdf_file.attrs[attr_name] = attr_val
            except TypeError as err:
                msg = f"Cannot store attribute {attr_name=} with value {attr_val=} in HDF."
                pylog.error(msg)
                attrs_errors.append(err)

    # Raises attributes errors after closing HDF file
    for err in attrs_errors:
        raise err

    if verbose >= 2:
        pylog.debug(f"Data has been packed into HDF file '{hdf_fpath}'.")

    if scan_results_fpath.is_file():
        os.remove(scan_results_fpath)

    hdf_dataset = HDFDataset(hdf_fpath, **ds_kwds)
    return hdf_dataset


def hdf_dtype_to_fill_value(hdf_dtype: Optional[HDFDType]) -> BuiltinScalar:
    if isinstance(hdf_dtype, np.dtype):
        hdf_dtype = hdf_dtype.type

    if hdf_dtype == "b" or hdf_dtype == np.bool_:
        return False
    elif hdf_dtype in ("i", "u") or (
        isinstance(hdf_dtype, type) and issubclass(hdf_dtype, np.integer)
    ):
        return 0
    elif hdf_dtype == "f" or (
        isinstance(hdf_dtype, type) and issubclass(hdf_dtype, np.floating)
    ):
        return 0.0
    elif (
        hdf_dtype == "c"
        or (
            isinstance(hdf_dtype, type)
            and (
                hdf_dtype in (np.void, np.object_, np.bytes_, np.str_)
                or issubclass(hdf_dtype, np.complexfloating)
            )
        )
        or (isinstance(hdf_dtype, np.dtype) and numpy_is_complex_dtype(hdf_dtype))
    ):
        return None
    else:
        msg = f"Unsupported type {hdf_dtype=}."
        raise ValueError(msg)


def numpy_dtype_to_hdf_dtype(
    dtype: Optional[np.dtype],
    *,
    encoding: str = HDF_ENCODING,
) -> np.dtype:
    if dtype is None:
        return HDF_VOID_DTYPE
    elif isinstance(dtype, np.dtype) and dtype.kind == "U":
        return h5py.string_dtype(encoding, None)
    else:
        return dtype


def hdf_dtype_to_numpy_dtype(hdf_dtype: HDFDType) -> np.dtype:
    if isinstance(hdf_dtype, np.dtype):
        return hdf_dtype
    if hdf_dtype == HDF_VOID_DTYPE:
        return np.dtype("V")
    if hdf_dtype == HDF_STRING_DTYPE:
        return np.dtype("<U")
    if hdf_dtype == "f":
        return np.dtype("float32")
    if hdf_dtype == "i":
        return np.dtype("int32")
    if hdf_dtype == "b":
        return np.dtype("int8")
    if hdf_dtype == "c":
        return np.dtype("|S1")

    raise ValueError(f"Unsupported dtype {hdf_dtype=} for numpy dtype.")


def _scan_dataset(
    dataset: SizedDatasetLike[T],
    pre_transform: Optional[Callable[[T], T_DictOrTuple]],
    batch_size: int,
    num_workers: int,
    store_str_as_vlen: bool,
    verbose: int,
    encoding: str,
    skip_scan: bool,
) -> Tuple[
    Callable[[T], Dict[str, Any]],
    HDFItemType,
    Dict[str, Tuple[int, ...]],
    Dict[str, HDFDType],
    Dict[str, bool],
    Dict[str, Set[np.dtype]],
]:
    if pre_transform is None:
        pre_transform = nn.Identity()

    if isinstance(dataset, IterableDataset):
        item_0 = next(iter(dataset))
    else:
        item_0 = dataset[0]

    def encode_array(x: np.ndarray) -> Any:
        if x.dtype.kind == "U":
            x = np.char.encode(x, encoding=encoding)
        if x.dtype.kind == "S":
            x = x.tolist()
        return x

    def encode_dict_array(x: Dict[str, np.ndarray]) -> Dict[str, Any]:
        return {k: encode_array(to.to_numpy(v)) for k, v in x.items()}  # type: ignore

    to_dict_fn: Callable[[T], Dict[str, Any]]

    if is_dict_str(item_0):
        item_type = "dict"
        to_dict_fn = to.identity  # type: ignore
    elif isinstance(item_0, tuple):
        item_type = "tuple"
        to_dict_fn = _tuple_to_dict  # type: ignore
    else:
        msg = f"Invalid item type for {dataset.__class__.__name__}. (expected Dict[str, Any] or tuple but found {type(item_0)})"
        raise ValueError(msg)
    del item_0

    encode_dict_fn = to.identity if store_str_as_vlen else encode_dict_array

    dict_pre_transform: Callable[[T], Dict[str, Any]] = Compose(
        pre_transform,
        to_dict_fn,
        encode_dict_fn,
    )

    loader = DataLoader(
        dataset,  # type: ignore
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        collate_fn=nn.Identity(),
        drop_last=False,
        pin_memory=False,
    )

    infos_dict: Dict[str, Set[Tuple[Tuple[int, ...], np.dtype]]] = {}
    src_np_dtypes: Dict[str, Set[np.dtype]] = {}

    for batch in tqdm.tqdm(
        loader,
        desc="Pre compute shapes...",
        disable=verbose <= 0 or skip_scan,
    ):
        batch = [pre_transform(item) for item in batch]
        batch = [to_dict_fn(item) for item in batch]

        for item in batch:
            for attr_name, value in item.items():
                info = scan_shape_dtypes(value, empty_np=HDF_VOID_DTYPE)
                shape = info.shape
                np_dtype = info.numpy_dtype
                kind = np_dtype.kind

                if attr_name in src_np_dtypes:
                    src_np_dtypes[attr_name].add(np_dtype)  # type: ignore
                else:
                    src_np_dtypes[attr_name] = {np_dtype}  # type: ignore

                value = to.to_numpy(value)
                if kind == "U" and not store_str_as_vlen:
                    value = encode_array(value)  # type: ignore
                    # update shape and np_dtype after encoding
                    info = scan_shape_dtypes(value, empty_np=HDF_VOID_DTYPE)
                    shape = info.shape
                    np_dtype = info.numpy_dtype

                if attr_name in infos_dict:
                    infos_dict[attr_name].add((shape, np_dtype))  # type: ignore
                else:
                    infos_dict[attr_name] = {(shape, np_dtype)}  # type: ignore

            if skip_scan:
                break

    max_shapes: Dict[str, Tuple[int, ...]] = {}
    hdf_dtypes: Dict[str, HDFDType] = {}
    all_eq_shapes: Dict[str, bool] = {}

    for attr_name, info in infos_dict.items():
        shapes = [shape for shape, _ in info]
        ndims = list(map(len, shapes))
        if not all_eq(ndims):
            ndims_set = tuple(set(ndims))
            indices = [ndims.index(ndim) for ndim in ndims_set]
            msg = f"Invalid ndim for attribute {attr_name}. (found multiple ndims: {ndims_set} at {indices=})"
            raise ValueError(msg)

        np_dtypes = [np_dtype for _, np_dtype in info]
        np_dtype = merge_numpy_dtypes(np_dtypes, empty=HDF_VOID_DTYPE)
        hdf_dtype = numpy_dtype_to_hdf_dtype(np_dtype, encoding=encoding)

        all_eq_shapes[attr_name] = all_eq(shapes)
        max_shapes[attr_name] = tuple(map(max, zip(*shapes)))
        hdf_dtypes[attr_name] = hdf_dtype

    del infos_dict

    if verbose >= 2:
        pylog.debug(f"Found max_shapes:\n{max_shapes}")
        pylog.debug(f"Found hdf_dtypes:\n{hdf_dtypes}")
        pylog.debug(f"Found all_eq_shapes:\n{all_eq_shapes}")
        pylog.debug(f"Found src_np_dtypes:\n{src_np_dtypes}")

    return (
        dict_pre_transform,
        item_type,
        max_shapes,
        hdf_dtypes,
        all_eq_shapes,
        src_np_dtypes,
    )


def bytearray_to_bytes(x: Any) -> Any:
    if isinstance(x, (str, bytes)):
        return x
    if isinstance(x, bytearray):
        return bytes(x)
    elif isinstance(x, Mapping):
        return {bytearray_to_bytes(k): bytearray_to_bytes(v) for k, v in x.items()}
    elif isinstance(x, Iterable):
        return type(x)(bytearray_to_bytes(xi) for xi in x)
    else:
        return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from .collate import AdvancedCollateDict, CollateDict
from .dataloader import get_auto_num_cpus, get_auto_num_gpus
from .dataset import (
    EmptyDataset,
    IterableSubset,
    IterableTransformWrapper,
    IterableWrapper,
    Subset,
    TransformWrapper,
    Wrapper,
)
from .slicer import DatasetSlicer, DatasetSlicerWrapper
from .split import balanced_monolabel_split, random_split



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.functional.multiclass import index_to_onehot, onehot_to_index


class TestMulticlass(TestCase):
    def test_index_to_onehot_1(self) -> None:
        indices = torch.as_tensor([[0, 2, 1], [0, 0, 2]])
        expected = torch.as_tensor(
            [[[1, 0, 0], [0, 0, 1], [0, 1, 0]], [[1, 0, 0], [1, 0, 0], [0, 0, 1]]],
            dtype=torch.bool,
        )
        result = index_to_onehot(indices, 3)
        assert torch.equal(result, expected)

    def test_index_to_onehot_2(self) -> None:
        indices = torch.as_tensor([[0, -1, 1], [0, 0, 2], [-1, -1, -1]])
        expected = torch.as_tensor(
            [
                [[1, 0, 0], [0, 0, 0], [0, 1, 0]],
                [[1, 0, 0], [1, 0, 0], [0, 0, 1]],
                [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
            ],
            dtype=torch.bool,
        )
        result = index_to_onehot(indices, 3, padding_idx=-1)
        assert torch.equal(result, expected)

    def test_onehot_to_index_3d(self) -> None:
        onehot = torch.as_tensor([[[0, 1, 0], [1, 0, 0]], [[0, 0, 1], [0, 0, 1]]])
        expected = torch.as_tensor([[1, 0], [2, 2]])
        result = onehot_to_index(onehot)
        assert torch.equal(result, expected)

    def test_index_to_onehot_3d(self) -> None:
        indices = [[1, 0], [2, 2]]
        expected = torch.as_tensor(
            [[[0, 1, 0], [1, 0, 0]], [[0, 0, 1], [0, 0, 1]]]
        ).bool()
        result = index_to_onehot(indices, 3)
        assert torch.equal(result, expected)

    def test_index_to_onehot_padding_idx(self) -> None:
        x = torch.randint(0, 100, (1000,))
        mask = torch.rand_like(x, dtype=torch.float) > 0.2
        x = torch.where(mask, -10, x)
        onehots = index_to_onehot(x, 100, padding_idx=-10)
        assert onehots[mask].eq(False).all()

        index = onehot_to_index(onehots, padding_idx=-10)
        assert torch.equal(x, index)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Iterable, List, Tuple

import torch
from torch import Tensor
from typing_extensions import TypeIs

from torchoutil.extras.numpy import is_numpy_number_like, is_numpy_scalar_like, np
from torchoutil.pyoutil.typing import is_builtin_number, is_builtin_scalar
from torchoutil.types._typing import (
    BoolTensor,
    BoolTensor1D,
    ComplexFloatingTensor,
    FloatingTensor,
    NumberLike,
    ScalarLike,
    SignedIntegerTensor,
    SignedIntegerTensor1D,
    Tensor0D,
    TensorLike,
)


def is_bool_tensor(x: Any) -> TypeIs[BoolTensor]:
    return isinstance(x, BoolTensor)


def is_bool_tensor1d(x: Any) -> TypeIs[BoolTensor1D]:
    return is_bool_tensor(x) and x.ndim == 1


def is_complex_tensor(x: Any) -> TypeIs[ComplexFloatingTensor]:
    return isinstance(x, Tensor) and x.is_complex()


def is_floating_tensor(x: Any) -> TypeIs[FloatingTensor]:
    return isinstance(x, Tensor) and x.is_floating_point()


def is_integral_dtype(dtype: torch.dtype) -> bool:
    return not dtype.is_floating_point and not dtype.is_complex and dtype.is_signed


def is_integral_tensor(x: Any) -> TypeIs[SignedIntegerTensor]:
    return isinstance(x, Tensor) and is_integral_dtype(x.dtype)


def is_integral_tensor1d(x: Any) -> TypeIs[SignedIntegerTensor1D]:
    return is_integral_tensor(x) and x.ndim == 1


def is_iterable_tensor(x: Any) -> TypeIs[Iterable[Tensor]]:
    return isinstance(x, Iterable) and all(isinstance(xi, Tensor) for xi in x)


def is_list_tensor(x: Any) -> TypeIs[List[Tensor]]:
    return isinstance(x, list) and all(isinstance(xi, Tensor) for xi in x)


def is_number_like(x: Any) -> TypeIs[NumberLike]:
    """Returns True if input is a scalar number.

    Accepted numbers-like objects are:
    - Python numbers (int, float, bool, complex)
    - Numpy zero-dimensional arrays
    - Numpy numbers
    - PyTorch zero-dimensional tensors
    """
    return is_builtin_number(x) or is_numpy_number_like(x) or is_tensor0d(x)


def is_scalar_like(x: Any) -> TypeIs[ScalarLike]:
    """Returns True if input is a scalar number.

    Accepted scalar-like objects are:
    - Python scalars like (int, float, bool, complex, None, str, bytes)
    - Numpy zero-dimensional arrays
    - Numpy generic
    - PyTorch zero-dimensional tensors
    """
    return is_builtin_scalar(x) or is_numpy_scalar_like(x) or is_tensor0d(x)


def is_tensor0d(x: Any) -> TypeIs[Tensor0D]:
    """Returns True if x is a zero-dimensional torch Tensor."""
    return isinstance(x, Tensor) and x.ndim == 0


def is_tensor_like(x: Any) -> TensorLike:
    return isinstance(x, (Tensor, np.ndarray))


def is_tuple_tensor(x: Any) -> TypeIs[Tuple[Tensor, ...]]:
    return isinstance(x, tuple) and all(isinstance(xi, Tensor) for xi in x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Any, Callable, Dict, Optional, TypeVar, Union

T = TypeVar("T", covariant=True)


def open_close_wrap(
    fn: Callable[..., T],
    fpath: Union[str, Path],
    open_kwds: Optional[Dict[str, Any]] = None,
    *args,
    **kwargs,
) -> T:
    if open_kwds is None:
        open_kwds = {}

    with open(fpath, **open_kwds) as file:
        return fn(file, *args, **kwargs)  # type: ignore



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import ClassVar, Dict, Type


class Singleton(type):
    """Singleton metaclass.

    To use it, just inherit from metaclass:
    ```
    >>> class MyClass(metaclass=Singleton):
    >>>     pass
    >>> a1 = MyClass()
    >>> a2 = MyClass()
    >>> # a1 and a2 are the same object, i.e. id(a1) == id(a2)
    ```
    """

    _instances: ClassVar[Dict[Type, object]] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        instance = cls._instances[cls]
        return instance



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from enum import Enum
from typing import Type, TypeVar

TStrEnum = TypeVar("TStrEnum", bound="StrEnum")


class StrEnum(str, Enum):
    @classmethod
    def from_str(
        cls: Type[TStrEnum],
        value: str,
        case_sensitive: bool = False,
    ) -> TStrEnum:
        members = cls.__members__.keys()
        for member in members:
            if member == value or (
                not case_sensitive and member.lower() == value.lower()
            ):
                return cls[member]

        msg = f"Invalid argument {value=}. (expected one of {tuple(members)})"
        raise ValueError(msg)

    @staticmethod
    def _generate_next_value_(name, start, count, last_values) -> str:
        return name

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        other = other.value if isinstance(other, Enum) else str(other)
        return self.value == other  # type: ignore

    def __hash__(self) -> int:
        return hash(self.value)
